import copy
from unittest import TestCase
from musicxml.xmlelement.containers import containers
from musicxml.xmlelement.xmlchildcontainer import XMLChildContainerFactory
from musicxml.xsd.xsdcomplextype import XSDComplexTypeNote


class TestPrecreatingXMLChildContainer(TestCase):
    def test_xml_note(self):
        container = copy.deepcopy(containers['XSDComplexTypeNote'])
        manually = XMLChildContainerFactory(XSDComplexTypeNote).get_child_container()
        assert manually.get_tree_representation() == container.get_tree_representation()



from musicxml.util.core import convert_to_xsd_class_name
from musicxml.xsd.xsdtree import XSDTree, XSDTreeElement, XSD_TREE_DICT
from musicxml.xsd.xsdsimpletype import *
import xml.etree.ElementTree as ET


class XSDAttribute:
    def __init__(self, xsd_tree):
        self._name = None
        self._ref = None
        self._xsd_tree = None
        self.xsd_tree = xsd_tree
        self._type = None
        self._is_required = None

    @property
    def xsd_tree(self):
        return self._xsd_tree

    @xsd_tree.setter
    def xsd_tree(self, value):
        if not isinstance(value, XSDTree):
            raise TypeError
        if value.tag != 'attribute':
            raise ValueError
        ref = value.get_attributes().get('ref')
        if ref:
            if ref == 'xml:lang':
                self._xsd_tree = XSDTree(ET.fromstring("""<xs:attribute xmlns:xs="http://www.w3.org/2001/XMLSchema" name="lang" type="xs:language">
        <xs:annotation>
            <xs:documentation>In due course, we should install the relevant ISO 2- and 3-letter
                codes as the enumerated possible values . . .
            </xs:documentation>
        </xs:annotation>
    </xs:attribute>
    """
                                                       ))
            elif ref == 'xml:space':
                self._xsd_tree = XSDTree(ET.fromstring("""<xs:attribute xmlns:xs="http://www.w3.org/2001/XMLSchema" name="space" default="preserve">
        <xs:simpleType>
            <xs:restriction base="xs:NCName">
                <xs:enumeration value="default"/>
                <xs:enumeration value="preserve"/>
            </xs:restriction>
        </xs:simpleType>
    </xs:attribute>
    """
                                                       ))
            else:
                NotImplementedError(ref)
        else:
            self._xsd_tree = value

    @property
    def name(self):
        if self._name is None:
            self._name = self.xsd_tree.get_attributes().get('name')
        return self._name

    @property
    def ref(self):
        if self._ref is None:
            self._ref = self.xsd_tree.get_attributes().get('ref')
        return self._ref

    @property
    def type_(self):
        if self._type is None:
            self._type = eval(convert_to_xsd_class_name(self.xsd_tree.get_attributes()['type'], 'simple_type'))
        return self._type

    @property
    def is_required(self):
        if self._is_required is None:
            if self.xsd_tree.get_attributes().get('use') == 'required':
                self._is_required = True
            else:
                self._is_required = False
        return self._is_required

    def __call__(self, value):
        return self.type_(value)

    def __str__(self):
        attrs = self.xsd_tree.get_attributes()
        return f"XSDAttribute{''.join([f'@{attribute}={self.xsd_tree.get_attributes()[attribute]}' for attribute in attrs])}"

    def __repr__(self):
        return self.__str__()


class XSDAttributeGroup(XSDTreeElement):
    _XSD_ATTRIBUTES = None

    @classmethod
    def get_xsd_attributes(cls):
        if cls._XSD_ATTRIBUTES is None:
            cls._XSD_ATTRIBUTES = []
            for child in cls.XSD_TREE.get_children():
                if child.tag == 'attribute':
                    cls._XSD_ATTRIBUTES.append(XSDAttribute(child))
                if child.tag == 'attributeGroup':
                    cls._XSD_ATTRIBUTES.extend(eval(child.xsd_element_class_name).get_xsd_attributes())
        return cls._XSD_ATTRIBUTES

# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_attributes.py
# -----------------------------------------------------


class XSDAttributeGroupBendSound(XSDAttributeGroup):
    """
    The bend-sound type is used for bend and slide elements, and is similar to the trill-sound attribute group. Here the beats element refers to the number of discrete elements (like MIDI pitch bends) used to represent a continuous bend or slide. The first-beat indicates the percentage of the duration for starting a bend; the last-beat the percentage for ending it. The default choices are:

	accelerate = "no"
	beats = "4"
	first-beat = "25"
	last-beat = "75"
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['bend-sound']


class XSDAttributeGroupBezier(XSDAttributeGroup):
    """
    The bezier attribute group is used to indicate the curvature of slurs and ties, representing the control points for a cubic bezier curve. For ties, the bezier attribute group is used with the tied element.

Normal slurs, S-shaped slurs, and ties need only two bezier points: one associated with the start of the slur or tie, the other with the stop. Complex slurs and slurs divided over system breaks can specify additional bezier data at slur elements with a continue type.

The bezier-x, bezier-y, and bezier-offset attributes describe the outgoing bezier point for slurs and ties with a start type, and the incoming bezier point for slurs and ties with types of stop or continue. The bezier-x2, bezier-y2, and bezier-offset2 attributes are only valid with slurs of type continue, and describe the outgoing bezier point.

The bezier-x, bezier-y, bezier-x2, and bezier-y2 attributes are specified in tenths, relative to any position settings associated with the slur or tied element. The bezier-offset and bezier-offset2 attributes are measured in terms of musical divisions, like the offset element. 

The bezier-offset and bezier-offset2 attributes are deprecated as of MusicXML 3.1. If both the bezier-x and bezier-offset attributes are present, the bezier-x attribute takes priority. Similarly, the bezier-x2 attribute takes priority over the bezier-offset2 attribute. The two types of bezier attributes are not additive.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['bezier']


class XSDAttributeGroupColor(XSDAttributeGroup):
    """
    The color attribute group indicates the color of an element.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['color']


class XSDAttributeGroupDashedFormatting(XSDAttributeGroup):
    """
    The dashed-formatting entity represents the length of dashes and spaces in a dashed line. Both the dash-length and space-length attributes are represented in tenths. These attributes are ignored if the corresponding line-type attribute is not dashed.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['dashed-formatting']


class XSDAttributeGroupDirective(XSDAttributeGroup):
    """
    The directive attribute changes the default-x position of a direction. It indicates that the left-hand side of the direction is aligned with the left-hand side of the time signature. If no time signature is present, it is aligned with the left-hand side of the first music notational element in the measure. If a default-x, justify, or halign attribute is present, it overrides the directive attribute.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['directive']


class XSDAttributeGroupDocumentAttributes(XSDAttributeGroup):
    """
    The document-attributes attribute group is used to specify the attributes for an entire MusicXML document. Currently this is used for the version attribute.

The version attribute was added in Version 1.1 for the score-partwise and score-timewise documents. It provides an easier way to get version information than through the MusicXML public ID. The default value is 1.0 to make it possible for programs that handle later versions to distinguish earlier version files reliably. Programs that write MusicXML 1.1 or later files should set this attribute.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['document-attributes']


class XSDAttributeGroupEnclosure(XSDAttributeGroup):
    """
    The enclosure attribute group is used to specify the formatting of an enclosure around text or symbols.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['enclosure']


class XSDAttributeGroupFont(XSDAttributeGroup):
    """
    The font attribute group gathers together attributes for determining the font within a credit or direction. They are based on the text styles for Cascading Style Sheets. The font-family is a comma-separated list of font names.The font-style can be normal or italic. The font-size can be one of the CSS sizes or a numeric point size. The font-weight can be normal or bold. The default is application-dependent, but is a text font vs. a music font.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['font']


class XSDAttributeGroupHalign(XSDAttributeGroup):
    """
    In cases where text extends over more than one line, horizontal alignment and justify values can be different. The most typical case is for credits, such as:

	Words and music by
	  Pat Songwriter

Typically this type of credit is aligned to the right, so that the position information refers to the right-most part of the text. But in this example, the text is center-justified, not right-justified.

The halign attribute is used in these situations. If it is not present, its value is the same as for the justify attribute. For elements where a justify attribute is not allowed, the default is implementation-dependent.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['halign']


class XSDAttributeGroupJustify(XSDAttributeGroup):
    """
    The justify attribute is used to indicate left, center, or right justification. The default value varies for different elements. For elements where the justify attribute is present but the halign attribute is not, the justify attribute indicates horizontal alignment as well as justification.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['justify']


class XSDAttributeGroupLetterSpacing(XSDAttributeGroup):
    """
    The letter-spacing attribute specifies text tracking. Values are either "normal" or a number representing the number of ems to add between each letter. The number may be negative in order to subtract space. The default is normal, which allows flexibility of letter-spacing for purposes of text justification.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['letter-spacing']


class XSDAttributeGroupLevelDisplay(XSDAttributeGroup):
    """
    The level-display attribute group specifies three common ways to indicate editorial indications: putting parentheses or square brackets around a symbol, or making the symbol a different size. If not specified, they are left to application defaults. It is used by the level and accidental elements.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['level-display']


class XSDAttributeGroupLineHeight(XSDAttributeGroup):
    """
    The line-height attribute specifies text leading. Values are either "normal" or a number representing the percentage of the current font height to use for leading. The default is "normal". The exact normal value is implementation-dependent, but values between 100 and 120 are recommended.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['line-height']


class XSDAttributeGroupLineLength(XSDAttributeGroup):
    """
    The line-length attribute distinguishes between different line lengths for doit, falloff, plop, and scoop articulations.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['line-length']


class XSDAttributeGroupLineShape(XSDAttributeGroup):
    """
    The line-shape attribute distinguishes between straight and curved lines.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['line-shape']


class XSDAttributeGroupLineType(XSDAttributeGroup):
    """
    The line-type attribute distinguishes between solid, dashed, dotted, and wavy lines.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['line-type']


class XSDAttributeGroupOptionalUniqueId(XSDAttributeGroup):
    """
    The optional-unique-id attribute group allows an element to optionally specify an ID that is unique to the entire document. This attribute group is not used for a required id attribute, or for an id attribute that specifies an id reference.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['optional-unique-id']


class XSDAttributeGroupOrientation(XSDAttributeGroup):
    """
    The orientation attribute indicates whether slurs and ties are overhand (tips down) or underhand (tips up). This is distinct from the placement attribute used by any notation type.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['orientation']


class XSDAttributeGroupPlacement(XSDAttributeGroup):
    """
    The placement attribute indicates whether something is above or below another element, such as a note or a notation.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['placement']


class XSDAttributeGroupPosition(XSDAttributeGroup):
    """
    For most elements, any program will compute a default x and y position. The position attributes let this be changed two ways.

The default-x and default-y attributes change the computation of the default position. For most elements, the origin is changed relative to the left-hand side of the note or the musical position within the bar (x) and the top line of the staff (y).

For the following elements, the default-x value changes the origin relative to the start of the current measure:

	- note
	- figured-bass
	- harmony
	- link
	- directive
	- measure-numbering
	- all descendants of the part-list element
	- all children of the direction-type element

This origin is from the start of the entire measure, at either the left barline or the start of the system.

When the default-x attribute is used within a child element of the part-name-display, part-abbreviation-display, group-name-display, or group-abbreviation-display elements, it changes the origin relative to the start of the first measure on the system. These values are used when the current measure or a succeeding measure starts a new system. The same change of origin is used for the group-symbol element.

For the note, figured-bass, and harmony elements, the default-x value is considered to have adjusted the musical position within the bar for its descendant elements.

Since the credit-words and credit-image elements are not related to a measure, in these cases the default-x and default-y attributes adjust the origin relative to the bottom left-hand corner of the specified page.

The relative-x and relative-y attributes change the position relative to the default position, either as computed by the individual program, or as overridden by the default-x and default-y attributes.

Positive x is right, negative x is left; positive y is up, negative y is down. All units are in tenths of interline space. For stems, positive relative-y lengthens a stem while negative relative-y shortens it.

The default-x and default-y position attributes provide higher-resolution positioning data than related features such as the placement attribute and the offset element. Applications reading a MusicXML file that can understand both features should generally rely on the default-x and default-y attributes for their greater accuracy. For the relative-x and relative-y attributes, the offset element, placement attribute, and directive attribute provide context for the relative position information, so the two features should be interpreted together.

As elsewhere in the MusicXML format, tenths are the global tenths defined by the scaling element, not the local tenths of a staff resized by the staff-size element.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['position']


class XSDAttributeGroupPrintObject(XSDAttributeGroup):
    """
    The print-object attribute specifies whether or not to print an object (e.g. a note or a rest). It is yes by default.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['print-object']


class XSDAttributeGroupPrintSpacing(XSDAttributeGroup):
    """
    The print-spacing attribute controls whether or not spacing is left for an invisible note or object. It is used only if no note, dot, or lyric is being printed. The value is yes (leave spacing) by default.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['print-spacing']


class XSDAttributeGroupPrintStyle(XSDAttributeGroup):
    """
    The print-style attribute group collects the most popular combination of printing attributes: position, font, and color.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['print-style']


class XSDAttributeGroupPrintStyleAlign(XSDAttributeGroup):
    """
    The print-style-align attribute group adds the halign and valign attributes to the position, font, and color attributes.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['print-style-align']


class XSDAttributeGroupPrintout(XSDAttributeGroup):
    """
    The printout attribute group collects the different controls over printing an object (e.g. a note or rest) and its parts, including augmentation dots and lyrics. This is especially useful for notes that overlap in different voices, or for chord sheets that contain lyrics and chords but no melody.

By default, all these attributes are set to yes. If print-object is set to no, the print-dot and print-lyric attributes are interpreted to also be set to no if they are not present.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['printout']


class XSDAttributeGroupSmufl(XSDAttributeGroup):
    """
    The smufl attribute group is used to indicate a particular Standard Music Font Layout (SMuFL) character. Sometimes this is a formatting choice, and sometimes this is a refinement of the semantic meaning of an element.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['smufl']


class XSDAttributeGroupSystemRelation(XSDAttributeGroup):
    """
    The system-relation attribute group distinguishes elements that are associated with a system rather than the particular part where the element appears.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['system-relation']


class XSDAttributeGroupSymbolFormatting(XSDAttributeGroup):
    """
    The symbol-formatting attribute group collects the common formatting attributes for musical symbols. Default values may differ across the elements that use this group.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['symbol-formatting']


class XSDAttributeGroupTextDecoration(XSDAttributeGroup):
    """
    The text-decoration attribute group is based on the similar feature in XHTML and CSS. It allows for text to be underlined, overlined, or struck-through. It extends the CSS version by allow double or triple lines instead of just being on or off.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['text-decoration']


class XSDAttributeGroupTextDirection(XSDAttributeGroup):
    """
    The text-direction attribute is used to adjust and override the Unicode bidirectional text algorithm, similar to the Directionality data category in the W3C Internationalization Tag Set recommendation.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['text-direction']


class XSDAttributeGroupTextFormatting(XSDAttributeGroup):
    """
    The text-formatting attribute group collects the common formatting attributes for text elements. Default values may differ across the elements that use this group.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['text-formatting']


class XSDAttributeGroupTextRotation(XSDAttributeGroup):
    """
    The rotation attribute is used to rotate text around the alignment point specified by the halign and valign attributes. Positive values are clockwise rotations, while negative values are counter-clockwise rotations.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['text-rotation']


class XSDAttributeGroupTrillSound(XSDAttributeGroup):
    """
    The trill-sound attribute group includes attributes used to guide the sound of trills, mordents, turns, shakes, and wavy lines. The default choices are:

	start-note = "upper"
	trill-step = "whole"
	two-note-turn = "none"
	accelerate = "no"
	beats = "4".

Second-beat and last-beat are percentages for landing on the indicated beat, with defaults of 25 and 75 respectively.

For mordent and inverted-mordent elements, the defaults are different:

	The default start-note is "main", not "upper".
	The default for beats is "3", not "4".
	The default for second-beat is "12", not "25".
	The default for last-beat is "24", not "75".
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['trill-sound']


class XSDAttributeGroupValign(XSDAttributeGroup):
    """
    The valign attribute is used to indicate vertical alignment to the top, middle, bottom, or baseline of the text. Defaults are implementation-dependent.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['valign']


class XSDAttributeGroupValignImage(XSDAttributeGroup):
    """
    The valign-image attribute is used to indicate vertical alignment for images and graphics, so it removes the baseline value. Defaults are implementation-dependent.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['valign-image']


class XSDAttributeGroupXPosition(XSDAttributeGroup):
    """
    The x-position attribute group is used for elements like notes where specifying x position is common, but specifying y position is rare.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['x-position']


class XSDAttributeGroupYPosition(XSDAttributeGroup):
    """
    The y-position attribute group is used for elements like stems where specifying y position is common, but specifying x position is rare.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['y-position']


class XSDAttributeGroupImageAttributes(XSDAttributeGroup):
    """
    The image-attributes group is used to include graphical images in a score. The required source attribute is the URL for the image file. The required type attribute is the MIME type for the image file format. Typical choices include application/postscript, image/gif, image/jpeg, image/png, and image/tiff. The optional height and width attributes are used to size and scale an image. The image should be scaled independently in X and Y if both height and width are specified. If only one attribute is specified, the image should be scaled proportionally to fit in the specified dimension.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['image-attributes']


class XSDAttributeGroupPrintAttributes(XSDAttributeGroup):
    """
    The print-attributes group is used by the print element. The new-system and new-page attributes indicate whether to force a system or page break, or to force the current music onto the same system or page as the preceding music. Normally this is the first music data within a measure. If used in multi-part music, they should be placed in the same positions within each part, or the results are undefined. The page-number attribute sets the number of a new page; it is ignored if new-page is not "yes". Version 2.0 adds a blank-page attribute. This is a positive integer value that specifies the number of blank pages to insert before the current measure. It is ignored if new-page is not "yes". These blank pages have no music, but may have text or images specified by the credit element. This is used to allow a combination of pages that are all text, or all text and images, together with pages of music.

The staff-spacing attribute specifies spacing between multiple staves in tenths of staff space. This is deprecated as of Version 1.1; the staff-layout element should be used instead. If both are present, the staff-layout values take priority.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['print-attributes']


class XSDAttributeGroupElementPosition(XSDAttributeGroup):
    """
    The element and position attributes are new as of Version 2.0. They allow for bookmarks and links to be positioned at higher resolution than the level of music-data elements. When no element and position attributes are present, the bookmark or link element refers to the next sibling element in the MusicXML file. The element attribute specifies an element type for a descendant of the next sibling element that is not a link or bookmark. The position attribute specifies the position of this descendant element, where the first position is 1. The position attribute is ignored if the element attribute is not present. For instance, an element value of "beam" and a position value of "2" defines the link or bookmark to refer to the second beam descendant of the next sibling element that is not a link or bookmark. This is equivalent to an XPath test of [.//beam[2]] done in the context of the sibling element.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['element-position']


class XSDAttributeGroupLinkAttributes(XSDAttributeGroup):
    """
    The link-attributes group includes all the simple XLink attributes supported in the MusicXML format. It is also used to connect a MusicXML score with MusicXML parts or a MusicXML opus.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['link-attributes']


class XSDAttributeGroupGroupNameText(XSDAttributeGroup):
    """
    The group-name-text attribute group is used by the group-name and group-abbreviation elements. The print-style and justify attribute groups are deprecated in MusicXML 2.0 in favor of the new group-name-display and group-abbreviation-display elements.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['group-name-text']


class XSDAttributeGroupMeasureAttributes(XSDAttributeGroup):
    """
    The measure-attributes group is used by the measure element. Measures have a required number attribute (going from partwise to timewise, measures are grouped via the number).

The implicit attribute is set to "yes" for measures where the measure number should never appear, such as pickup measures and the last half of mid-measure repeats. The value is "no" if not specified.

The non-controlling attribute is intended for use in multimetric music like the Don Giovanni minuet. If set to "yes", the left barline in this measure does not coincide with the left barline of measures in other parts. The value is "no" if not specified.

In partwise files, the number attribute should be the same for measures in different parts that share the same left barline. While the number attribute is often numeric, it does not have to be. Non-numeric values are typically used together with the implicit or non-controlling attributes being set to "yes". For a pickup measure, the number attribute is typically set to "0" and the implicit attribute is typically set to "yes". 

If measure numbers are not unique within a part, this can cause problems for conversions between partwise and timewise formats. The text attribute allows specification of displayed measure numbers that are different than what is used in the number attribute. This attribute is ignored for measures where the implicit attribute is set to "yes". Further details about measure numbering can be specified using the measure-numbering element.

Measure width is specified in tenths. These are the global tenths specified in the scaling element, not local tenths as modified by the staff-size element.	The width covers the entire measure from barline or system start to barline or system end.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['measure-attributes']


class XSDAttributeGroupPartAttributes(XSDAttributeGroup):
    """
    In either partwise or timewise format, the part element has an id attribute that is an IDREF back to a score-part in the part-list.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['part-attributes']


class XSDAttributeGroupPartNameText(XSDAttributeGroup):
    """
    The part-name-text attribute group is used by the part-name and part-abbreviation elements. The print-style and justify attribute groups are deprecated in MusicXML 2.0 in favor of the new part-name-display and part-abbreviation-display elements.
    """
    XSD_TREE = XSD_TREE_DICT['attributeGroup']['part-name-text']

__all__=['XSDAttribute', 'XSDAttributeGroup', 'XSDAttributeGroupBendSound', 'XSDAttributeGroupBezier', 'XSDAttributeGroupColor', 'XSDAttributeGroupDashedFormatting', 'XSDAttributeGroupDirective', 'XSDAttributeGroupDocumentAttributes', 'XSDAttributeGroupEnclosure', 'XSDAttributeGroupFont', 'XSDAttributeGroupHalign', 'XSDAttributeGroupJustify', 'XSDAttributeGroupLetterSpacing', 'XSDAttributeGroupLevelDisplay', 'XSDAttributeGroupLineHeight', 'XSDAttributeGroupLineLength', 'XSDAttributeGroupLineShape', 'XSDAttributeGroupLineType', 'XSDAttributeGroupOptionalUniqueId', 'XSDAttributeGroupOrientation', 'XSDAttributeGroupPlacement', 'XSDAttributeGroupPosition', 'XSDAttributeGroupPrintObject', 'XSDAttributeGroupPrintSpacing', 'XSDAttributeGroupPrintStyle', 'XSDAttributeGroupPrintStyleAlign', 'XSDAttributeGroupPrintout', 'XSDAttributeGroupSmufl', 'XSDAttributeGroupSystemRelation', 'XSDAttributeGroupSymbolFormatting', 'XSDAttributeGroupTextDecoration', 'XSDAttributeGroupTextDirection', 'XSDAttributeGroupTextFormatting', 'XSDAttributeGroupTextRotation', 'XSDAttributeGroupTrillSound', 'XSDAttributeGroupValign', 'XSDAttributeGroupValignImage', 'XSDAttributeGroupXPosition', 'XSDAttributeGroupYPosition', 'XSDAttributeGroupImageAttributes', 'XSDAttributeGroupPrintAttributes', 'XSDAttributeGroupElementPosition', 'XSDAttributeGroupLinkAttributes', 'XSDAttributeGroupGroupNameText', 'XSDAttributeGroupMeasureAttributes', 'XSDAttributeGroupPartAttributes', 'XSDAttributeGroupPartNameText']



class XSDException(Exception):
    pass


class XSDAttributeRequiredException(XSDException):
    pass


class XSDWrongAttribute(XSDAttributeRequiredException):
    pass


class XMLElementException(Exception):
    pass


class XMLElementWrongChildType(XMLElementException):
    pass


class XMLElementChildrenRequired(XMLElementException):
    pass


from musicxml.util.core import cap_first, convert_to_xml_class_name
from musicxml.xsd.xsdtree import XSDTree, XSDTreeElement, XSD_TREE_DICT
import xml.etree.ElementTree as ET


class XSDSequence:
    def __init__(self, xsd_tree):
        self._xsd_tree = None
        self._elements = None
        self.xsd_tree = xsd_tree

    @property
    def elements(self):
        if not self._elements:
            self._elements = []
            for child in self.xsd_tree.get_children():
                if child.tag == 'element':
                    element = convert_to_xml_class_name(child.name)
                    min_occurrence = child.get_attributes().get('minOccurs')
                    if min_occurrence is None: min_occurrence = '1'
                    max_occurrence = child.get_attributes().get('maxOccurs')
                    if max_occurrence is None: max_occurrence = '1'
                    self._elements.append((element, min_occurrence, max_occurrence))

                elif child.tag == 'group':
                    xsd_group_name = 'XSDGroup' + ''.join([cap_first(partial) for partial in child.get_attributes()['ref'].split('-')])
                    elements = eval(xsd_group_name)().sequence.elements
                    min_occurrence = child.get_attributes().get('minOccurs')
                    max_occurrence = child.get_attributes().get('maxOccurs')
                    if min_occurrence is not None:
                        if len(elements) > 1:
                            raise NotImplementedError
                        list_el = list(elements[0])
                        list_el[1] = min_occurrence
                        elements[0] = tuple(list_el)
                    if max_occurrence is not None:
                        if len(elements) > 1:
                            raise NotImplementedError
                        list_el = list(elements[0])
                        list_el[2] = max_occurrence
                        elements[0] = tuple(list_el)
                    self._elements.extend(elements)
                else:
                    raise NotImplementedError(child.tag)
        return self._elements

    @property
    def xsd_tree(self):
        return self._xsd_tree

    @xsd_tree.setter
    def xsd_tree(self, value):
        if not isinstance(value, XSDTree):
            raise TypeError
        if value.tag != 'sequence':
            raise ValueError
        self._xsd_tree = value

    def __copy__(self):
        return self.__class__(xsd_tree=self.xsd_tree)


class XSDChoice:
    def __init__(self, xsd_tree):
        self._xsd_tree = None
        self.xsd_tree = xsd_tree

    @property
    def xsd_tree(self):
        return self._xsd_tree

    @xsd_tree.setter
    def xsd_tree(self, value):
        if not isinstance(value, XSDTree):
            raise TypeError
        if value.tag != 'choice':
            raise ValueError
        self._xsd_tree = value

    def __copy__(self):
        return self.__class__(xsd_tree=self.xsd_tree)


class XSDGroup(XSDTreeElement):

    def __init__(self):
        self._sequence = None

    @property
    def name(self):
        return self.XSD_TREE.name

    @property
    def sequence(self):
        if not self._sequence:
            for child in self.XSD_TREE.get_children():
                if child.tag == 'sequence':
                    self._sequence = XSDSequence(child)
        return self._sequence

    @property
    def xsd_tree(self):
        return self.XSD_TREE

    def __copy__(self):
        copied = self.__class__()
        copied._sequence = self.sequence
        copied.XSD_TREE = self.XSD_TREE
        return copied
# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_indicators.py
# -----------------------------------------------------



from unittest import TestCase

from musicxml.exceptions import XMLElementChildrenRequired, XSDAttributeRequiredException, XSDWrongAttribute
from musicxml import *
from musicxml.xsd.xsdcomplextype import *
from musicxml.xsd.xsdsimpletype import *


class TestXMLElements(TestCase):

    def test_element_type(self):
        el = XMLOffset(2)
        assert el.TYPE == XSDComplexTypeOffset

        el = XMLElevation(2)
        assert el.TYPE == XSDSimpleTypeRotationDegrees

    def test_element_simple_content(self):
        """
        Test if complex types with a simple context (extension of a simple type) work properly in an XMLElement.
        A simple example is

        complexType@name=offset

        simpleContent
            extension@base=divisions
                attribute@name=sound@type=yes-no
        """
        el = XMLOffset(-2)
        assert el.to_string() == '<offset>-2</offset>\n'

        el = XMLOffset(-2, sound='yes')
        assert el.to_string() == '<offset sound="yes">-2</offset>\n'

        el = XMLOffset(2)
        # with self.assertRaises(XMLElementValueRequiredError):
        #     el.to_string()

        with self.assertRaises(TypeError):
            XMLOffset('wrong', sound='yes')

        with self.assertRaises(TypeError):
            XMLOffset(-2, sound=3).to_string()

        with self.assertRaises(ValueError):
            XMLOffset(-2, sound='maybe').to_string()

    def test_element_name(self):
        el = XMLOffset(-2)
        assert el.name == 'offset'
        el = XMLElevation(10)
        assert el.name == 'elevation'

    def test_element_with_wrong_attribute(self):
        with self.assertRaises(XSDWrongAttribute):
            XMLPartName('Part 1', dummy='no')
        el = XMLPartName('Part 1')
        with self.assertRaises(AttributeError):
            el.dummy = 'no'

    def test_xml_measure_attributes(self):
        m = XMLMeasure(number='2')
        m.text = 'hello'
        m.width = 150
        assert (m.number, m.text, m.width) == ('2', 'hello', 150)

    def test_xml_accidental_attributes(self):
        """
            complexType@name=accidental
        annotation
            documentation
        simpleContent
            extension@base=accidental-value
                attribute@name=cautionary@type=yes-no
                attribute@name=editorial@type=yes-no
                attributeGroup@ref=level-display
                attributeGroup@ref=print-style
                attribute@name=smufl@type=smufl-accidental-glyph-name
        :return:
        """
        a = XMLAccidental('sharp')
        a.cautionary = 'yes'
        assert a.cautionary == 'yes'

    def test_element_attribute_with_hyphenated_name(self):
        el = XMLPartName('Part 1', default_x=10)
        assert el.to_string() == """<part-name default-x="10">Part 1</part-name>
"""
        el.print_object = 'no'
        assert el.to_string() == """<part-name default-x="10" print-object="no">Part 1</part-name>
"""
        el = XMLPartName('Part 1', print_object='no')
        assert el.to_string() == """<part-name print-object="no">Part 1</part-name>
"""

    def test_element_with_simple_type(self):
        """
        <xs:element name="elevation" type="rotation-degrees" minOccurs="0">
            <xs:annotation>
                <xs:documentation>The elevation and pan elements allow placing of sound in a 3-D space relative to the listener. Both are expressed in degrees ranging from -180 to 180. For elevation, 0 is level with the listener, 90 is directly above, and -90 is directly below.</xs:documentation>
            </xs:annotation>
        </xs:element>
        """
        el = XMLElevation(10)
        assert el.TYPE == XSDSimpleTypeRotationDegrees
        with self.assertRaises(TypeError):
            el.value_ = 'something'
        with self.assertRaises(ValueError):
            el.value_ = 200

        el.value_ = 170
        assert el.to_string() == '<elevation>170</elevation>\n'

    def test_element_empty(self):
        """
        Test that empty complex type is created properly
        """
        el = XMLChord()
        assert el.to_string() == '<chord />\n'

    def test_sequence_indicator_children_required(self):
        """
        Test that a sequence indicator with only elements as children can verify the behavior of its corresponding element
        """
        """
        complexType@name=pitch
        sequence
            element@name=step@type=step
            element@name=alter@type=semitones@minOccurs=0
            element@name=octave@type=octave
        """

        """
        Element Pitch must have one and only one child element step and one and only one child element octave. It can have only one child 
        alter. The sequence order will be automatically set according to the sequence (step, alter, octave)
        """

        el = XMLPitch()
        with self.assertRaises(XMLElementChildrenRequired):
            el.to_string()

    def test_sequence_add_children_to_string(self):
        """
        Test that to_string function of an element with complex type and sequence works properly
        """
        el = XMLPitch()
        el.add_child(XMLStep('A'))
        el.add_child(XMLOctave(4))
        expected = """<pitch>
  <step>A</step>
  <octave>4</octave>
</pitch>
"""
        assert el.to_string() == expected

    def test_xml_element_part_list(self):
        el = XMLPartList()
        with self.assertRaises(XMLElementChildrenRequired) as err:
            el.to_string()

        assert err.exception.args[0] == 'XMLPartList requires at least following children: XMLScorePart'
        sp = el.add_child(XMLScorePart())

        with self.assertRaises(XMLElementChildrenRequired) as err:
            el.to_string()
        assert err.exception.args[0] == 'XMLScorePart requires at least following children: XMLPartName'

        pn = sp.add_child(XMLPartName('part name 1'))
        with self.assertRaises(XSDAttributeRequiredException) as err:
            el.to_string()
        assert err.exception.args[0] == 'XMLScorePart requires attribute: id'
        sp.id = 'p1'
        expected = """<part-list>
  <score-part id="p1">
    <part-name>part name 1</part-name>
  </score-part>
</part-list>
"""
        assert el.to_string() == expected

    def test_order_of_children(self):
        el = XMLAttributes()
        t = el.add_child(XMLTime())
        t.add_child(XMLBeats('4'))
        t.add_child(XMLBeatType('4'))
        el.add_child(XMLDivisions(1))
        c = el.add_child(XMLClef())
        c.add_child(XMLSign('G'))
        c.add_child(XMLLine(2))
        k = el.add_child(XMLKey())
        k.add_child(XMLFifths(0))
        k.add_child(XMLMode('major'))

        expected = """<attributes>
  <divisions>1</divisions>
  <key>
    <fifths>0</fifths>
    <mode>major</mode>
  </key>
  <time>
    <beats>4</beats>
    <beat-type>4</beat-type>
  </time>
  <clef>
    <sign>G</sign>
    <line>2</line>
  </clef>
</attributes>
"""
        assert el.to_string() == expected

    def test_element_with_xs_simple_type(self):
        """
        Test if there is no conflict between simple and complex types with the same name such as:
        xs:string vs. string: xs:string is a simple type, string is a complex type.
        """
        string = XMLString(2)
        assert string.TYPE == XSDComplexTypeString
        software = XMLSoftware('gaga')
        assert software.TYPE == XSDSimpleTypeString

    def test_element_with_type_and_value_as_attributes(self):
        """
        Test if an element which has attributes name type and value can be initiated without conflicts
        """
        supports = XMLSupports()
        assert supports.TYPE == XSDComplexTypeSupports
        supports.element = 'print'
        supports.type = 'yes'
        supports.value = 'yes'

    def test_xml_credit_words_attributes(self):
        """
        Test that XMLCreditWords can have attributes
        """
        cw = XMLCreditWords('something')
        assert [a.name for a in cw.TYPE.get_xsd_attributes()] == ['justify', 'default-x', 'default-y', 'relative-x',
                                                                  'relative-y',
                                                                  'font-family', 'font-style', 'font-size',
                                                                  'font-weight', 'color',
                                                                  'halign', 'valign', 'underline', 'overline',
                                                                  'line-through', 'rotation',
                                                                  'letter-spacing', 'line-height', 'lang', 'space',
                                                                  'dir', 'enclosure',
                                                                  'id']

    def test_xml_element_directive(self):
        d = XMLDirective('HU')
        d.lang = 'en'
        assert d.to_string() == '<directive lang="en">HU</directive>\n'

    def test_xml_note_with_tie(self):
        expected = """<note>
  <pitch>
    <step>F</step>
    <alter>1</alter>
    <octave>5</octave>
  </pitch>
  <duration>192</duration>
  <tie type="start" />
  <voice>1</voice>
  <type>half</type>
  <stem default-y="28">up</stem>
  <staff>1</staff>
  <notations>
    <tied orientation="over" type="start" />
  </notations>
</note>
"""
        n = XMLNote()
        p = n.add_child(XMLPitch())
        p.add_child(XMLStep('F'))
        p.add_child(XMLAlter(1))
        p.add_child(XMLOctave(5))
        n.add_child(XMLDuration(192))
        n.add_child(XMLTie(type='start'))
        n.add_child(XMLVoice('1'))
        n.add_child(XMLType('half'))
        n.add_child(XMLStem('up', default_y=28))
        n.add_child(XMLStaff(1))
        nn = n.add_child(XMLNotations())
        nn.add_child(XMLTied(orientation='over', type='start'))
        assert n.to_string() == expected

    def test_check_required_attributes(self):
        """
        Test if xml element's _check_required_attributes works properly.
        """
        br = XMLBeatRepeat()
        with self.assertRaises(XSDAttributeRequiredException):
            br._check_required_attributes()
        br.type = 'start'
        br._check_required_attributes()

    def test_font(self):
        expected = """<words default-y="-29" font-family="Arial" font-size="3" relative-x="-38">/</words>
"""
        w = XMLWords('/', default_y=-29, font_family="Arial", font_size=3, relative_x=-38)
        assert w.to_string() == expected
        expected = """<music-font font-family="Maestro,engraved" font-size="18.2" />
"""
        with self.assertRaises(ValueError):
            XMLMusicFont(font_family="Maestro,engraved", font_size="18.2")

        mf = XMLMusicFont(font_family="Maestro,engraved", font_size=18.2)
        assert mf.to_string() == expected

    def test_xml_credit(self):
        expected = """<credit page="1">
  <credit-words default-x="651" default-y="88" font-size="10" justify="center" valign="bottom">#</credit-words>
</credit>
"""
        c = XMLCredit(page=1)
        c.child_container_tree.check_requirements()
        c.add_child(XMLCreditWords('#', default_x=651, default_y=88, font_size=10, justify='center', valign='bottom'))
        assert c.to_string() == expected

    def test_xml_time_modification(self):
        expected = """<time-modification>
  <actual-notes>3</actual-notes>
  <normal-notes>2</normal-notes>
</time-modification>
"""
        tm = XMLTimeModification()
        tm.add_child(XMLActualNotes(3))
        tm.add_child(XMLNormalNotes(2))
        assert tm.to_string() == expected

    def test_xml_ending(self):
        expected = """<ending default-y="40" end-length="30" font-size="7.6" number="1" print-object="yes" type="start" />
"""
        e = XMLEnding(default_y=40, end_length=30, font_size=7.6, number='1', print_object='yes', type='start',
                      value_="")
        assert e.to_string() == expected

    def test_get_unordered_children(self):
        """
        Test XMLElement.get_children(ordered=False) which returns a list of elements in order of their addition
        """
        n = XMLNote()
        n.add_child(XMLType('half'))
        n.add_child(XMLDuration(192))
        p = n.add_child(XMLPitch())
        p.add_child(XMLStep('F'))
        p.add_child(XMLAlter(1))
        p.add_child(XMLOctave(5))
        n.add_child(XMLTie(type='start'))
        n.add_child(XMLVoice('1'))

        n.add_child(XMLStem('up', default_y=28))
        n.add_child(XMLStaff(1))
        nn = n.add_child(XMLNotations())
        nn.add_child(XMLTied(orientation='over', type='start'))

        assert [el.__class__.__name__ for el in n.get_children(ordered=False)] == ['XMLType', 'XMLDuration', 'XMLPitch',
                                                                                   'XMLTie',
                                                                                   'XMLVoice',
                                                                                   'XMLStem', 'XMLStaff',
                                                                                   'XMLNotations']

    def test_find_children(self):
        """
        Test XMLElement.find_child or find_children (ordered=False)
        """
        """
        complexType@name=appearance
            annotation
                documentation
            sequence
                element@name=line-width@type=line-width@minOccurs=0@maxOccurs=unbounded
                element@name=note-size@type=note-size@minOccurs=0@maxOccurs=unbounded
                element@name=distance@type=distance@minOccurs=0@maxOccurs=unbounded
                element@name=glyph@type=glyph@minOccurs=0@maxOccurs=unbounded
                element@name=other-appearance@type=other-appearance@minOccurs=0@maxOccurs=unbounded
        """
        n = XMLAppearance()
        ns1 = n.add_child(XMLNoteSize(10))
        lw = n.add_child(XMLLineWidth(10))
        ns2 = n.add_child(XMLNoteSize(10))
        assert n.find_child(XMLLineWidth) == lw
        assert n.find_child('XMLLineWidth') == lw
        assert n.find_children(XMLNoteSize) == [ns1, ns2]
        assert n.find_children('XMLNoteSize') == [ns1, ns2]

    def test_barline(self):
        b = XMLBarline()
        b.add_child(XMLBarStyle('light-light'))
        expected = """<barline>
  <bar-style>light-light</bar-style>
</barline>
"""
        assert b.to_string() == expected

    def test_possible_children_names(self):
        n = XMLNote()
        assert n.possible_children_names == {'beam', 'footnote', 'accidental', 'rest', 'unpitched', 'dot',
                                             'time-modification', 'tie',
                                             'instrument', 'level', 'notehead', 'duration', 'voice', 'stem', 'chord',
                                             'grace',
                                             'pitch', 'lyric', 'staff', 'listen', 'notations', 'type', 'cue', 'play',
                                             'notehead-text'}

    def test_convert_attribute_to_child(self):
        """
        Tests if a dot operator can create and add child if necessary
        """
        b = XMLBarline()
        assert b.xml_bar_style is None
        b.xml_bar_style = 'light-light'
        expected = """<barline>
  <bar-style>light-light</bar-style>
</barline>
"""
        assert b.to_string() == expected
        assert isinstance(b.xml_bar_style, XMLBarStyle)
        assert b.xml_bar_style.value_ == 'light-light'
        current_xml_bar_style = b.xml_bar_style

        b.xml_bar_style = 'light-heavy'
        assert b.xml_bar_style == current_xml_bar_style
        assert b.xml_bar_style.value_ == 'light-heavy'
        b.xml_bar_style = XMLBarStyle('regular')
        assert b.xml_bar_style.value_ == 'regular'
        assert b.xml_bar_style != current_xml_bar_style
        m = XMLMeasure()
        assert m.attributes == {}

        m = XMLMeasure(number='1')
        m.xml_barline = XMLBarline()

        m.xml_attributes = XMLAttributes()
        assert m.attributes == {'number': '1'}
        with self.assertRaises(AttributeError):
            m.barline
        m.xml_attributes.xml_divisions = 12
        m.xml_barline.xml_bar_style = 'light-light'

        expected = """<measure number="1">
  <barline>
    <bar-style>light-light</bar-style>
  </barline>
  <attributes>
    <divisions>12</divisions>
  </attributes>
</measure>
"""
        assert m.to_string() == expected

    def test_remove_child(self):
        """
        Test if element's can be removed
        """
        b = XMLBarline()
        b.xml_bar_style = 'light-light'
        expected = """<barline>
  <bar-style>light-light</bar-style>
</barline>
"""
        assert b.to_string() == expected
        b.remove(b.xml_bar_style)
        expected = """<barline />
"""
        assert b.to_string() == expected
        assert b.xml_bar_style is None
        b.xml_bar_style = 'light-light'
        assert b.xml_bar_style is not None
        b.xml_bar_style = None
        assert b.xml_bar_style is None
        assert b.get_children() == []
        expected = """<barline />
"""
        assert b.to_string() == expected

    def test_change_children(self):
        """
        Test if changing children directly influences parent's et_xml_element
        """
        n = XMLNote()
        n.xml_duration = 2
        p = n.xml_pitch = XMLPitch()
        p.xml_step = 'G'
        p.xml_alter = -1
        p.xml_octave = 4
        expected = """<note>
  <pitch>
    <step>G</step>
    <alter>-1</alter>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
</note>
"""
        assert n.to_string() == expected
        p.xml_alter = None
        expected = """<note>
  <pitch>
    <step>G</step>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
</note>
"""
        assert n.to_string() == expected

    def test_remove_choices_chosen_element(self):
        n = XMLNote()
        n.xml_duration = 2
        p = n.xml_pitch = XMLPitch()
        p.xml_step = 'G'
        p.xml_alter = -1
        p.xml_octave = 4
        choice_container = p.parent_xsd_element.parent_container.get_parent()
        assert choice_container.chosen_child == p.parent_xsd_element.parent_container
        n.remove(n.xml_pitch)
        assert choice_container.chosen_child is None

    def test_remove_pitch_add_rest(self):
        n = XMLNote()
        n.xml_duration = 2
        p = n.xml_pitch = XMLPitch()
        p.xml_step = 'G'
        p.xml_alter = -1
        p.xml_octave = 4
        n.remove(n.xml_pitch)
        n.add_child(XMLRest())
        expected = """<note>
  <rest />
  <duration>2</duration>
</note>
"""
        assert n.to_string() == expected

    def test_remove_element_and_check_child_container(self):
        """
        Test if after removing an element roots container is checked again.
        """
        n = XMLNote()
        n.xml_duration = 1
        with self.assertRaises(XMLElementChildrenRequired):
            n.to_string()
        n.xml_rest = XMLRest()
        n.to_string()
        n.remove(n.xml_rest)
        with self.assertRaises(XMLElementChildrenRequired):
            n.to_string()

    def test_attributes_underline(self):
        """
        Test if note can set and get attributes via dot operator
        """
        n = XMLNote()
        n.xml_rest = XMLRest()
        n.xml_duration = 2

        assert n.attack is None
        n.attack = 10
        assert n.attack == 10
        assert n.print_leger is None
        n.print_leger = 'yes'
        assert n.print_leger == 'yes'
        n.relative_x = 10
        assert n.relative_x == 10
        expected = """<note attack="10" print-leger="yes" relative-x="10">
  <rest />
  <duration>2</duration>
</note>
"""
        assert n.to_string() == expected

    def test_attributes_none(self):
        n = XMLNote(relative_x=10)
        n.xml_rest = XMLRest()
        n.xml_duration = 2
        n.print_leger = 'yes'
        n.relative_x = 10
        n.print_leger = None
        expected = """<note relative-x="10">
  <rest />
  <duration>2</duration>
</note>
"""
        assert n.to_string() == expected

    def test_xml_type_restrictions(self):
        with self.assertRaises(ValueError):
            XSDSimpleTypeNoteTypeValue('bla')
        with self.assertRaises(ValueError):
            XSDComplexTypeNoteType('bla')
        with self.assertRaises(ValueError):
            XMLType('bla')
        t = XMLType('whole')
        expected = """<type>whole</type>
"""
        assert t.to_string() == expected

    def test_xml_with_complex_content(self):
        """
        Test if value of a complex type is checked according to the core complex content
        """
        mt = XMLMetronomeTuplet()
        with self.assertRaises(XMLElementChildrenRequired):
            mt.to_string()
        mt.xml_actual_notes = 3
        mt.xml_normal_notes = 2
        with self.assertRaises(XSDAttributeRequiredException):
            mt.to_string()
        mt.type = 'start'
        expected = """<metronome-tuplet type="start">
  <actual-notes>3</actual-notes>
  <normal-notes>2</normal-notes>
</metronome-tuplet>
"""
        assert mt.to_string() == expected

    def test_xml_child_get_parent(self):
        mt = XMLMetronomeTuplet()
        an = mt.add_child(XMLActualNotes(3))
        mt.xml_normal_notes = 2
        assert an.get_parent() == mt
        assert mt.xml_normal_notes.get_parent() == mt
        assert mt.xml_normal_notes.up == mt

    def test_xml_remove_one_child_of_multiple_occurs(self):
        t = XMLTime()
        t.add_child(XMLBeats('3'))
        t.add_child(XMLBeatType('4'))
        t.add_child(XMLBeats('2'))
        t.add_child(XMLBeatType('4'))
        expected = """<time>
  <beats>3</beats>
  <beat-type>4</beat-type>
  <beats>2</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert t.to_string() == expected
        t.remove(t.find_children("XMLBeats")[1])
        t.remove(t.find_children("XMLBeatType")[1])
        expected = """<time>
  <beats>3</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert t.to_string() == expected
        t.add_child(XMLBeats('2'))
        t.add_child(XMLBeatType('4'))
        t.remove(t.find_children("XMLBeats")[0])
        t.remove(t.find_children("XMLBeatType")[0])
        expected = """<time>
  <beats>2</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert t.to_string() == expected

    def test_set_value_to_None(self):
        st = XMLStaff(1)
        assert st.value_ == 1
        st.value_ = 2
        assert st.value_ == 2
        st.value_ = 3
        assert st.value_ == 3
        expected = """<staff>3</staff>
"""
        assert st.to_string() == expected

    def test_xml_notations(self):
        n = XMLNotations()
        t1 = n.add_child(XMLTied(type='stop'))
        t2 = n.add_child(XMLTied(type='start'))
        expected = """<notations>
  <tied type="stop" />
  <tied type="start" />
</notations>
"""
        assert n.to_string() == expected
        t1.up.remove(t1)
        expected = """<notations>
  <tied type="start" />
</notations>
"""
        assert n.to_string() == expected

    def test_xml_articulations(self):
        n = XMLNotations()
        a = n.add_child(XMLArticulations())
        a.add_child(XMLAccent())
        a.add_child(XMLStaccato())
        expected = """<notations>
  <articulations>
    <accent />
    <staccato />
  </articulations>
</notations>
"""
        assert n.to_string() == expected

    def test_xml_fret(self):
        """
        Test if a value is required
        """
        with self.assertRaises(TypeError):
            f = XMLFret()
        f = XMLFret(value_=2)
        expected = """<fret>2</fret>
"""
        assert f.to_string() == expected

    def test_xml_xsd_check_false(self):
        """
        Test if xsd checking can be turned off.
        """
        p = XMLPart(id='p1', xsd_check=False)
        p.add_child(XMLTie())
        p.add_child(XMLKey())
        expected = """<part id="p1">
  <tie />
  <key />
</part>
"""
        assert p.to_string() == expected

    def test_xml_xsd_check_false_remove_child(self):
        """
        Test if removing child can take place if xsd checking is False.
        """
        p = XMLPart(id='p1', xsd_check=False)
        t = p.add_child(XMLTie())
        p.add_child(XMLKey())
        p.remove(t)
        expected = """<part id="p1">
  <key />
</part>
"""
        assert p.to_string() == expected

    def test_xml_xsd_check_false_replace_child(self):
        """
        Test if replacing child can take place if xsd checking is False.
        """
        p = XMLPart(id='p1', xsd_check=False)
        t = p.add_child(XMLTie())
        p.add_child(XMLKey())
        time = p.replace_child(t, XMLTime())
        time.add_child(XMLSenzaMisura())
        expected = """<part id="p1">
  <time>
    <senza-misura />
  </time>
  <key />
</part>
"""
        assert p.to_string() == expected

    def test_xml_ornaments(self):
        ornaments = XMLOrnaments()
        ornaments.add_child(XMLTrillMark(default_y=10))
        ornaments.add_child(XMLWavyLine(number=1, type='start', default_y=10))
        ornaments.add_child(XMLWavyLine(number=1, type='stop', relative_x=81))
        expected = """<ornaments>
  <trill-mark default-y="10" />
  <wavy-line number="1" type="start" default-y="10" />
  <wavy-line number="1" type="stop" relative-x="81" />
</ornaments>
"""
        assert ornaments.to_string() == expected



from unittest import TestCase

from musicxml.exceptions import XMLElementChildrenRequired
from musicxml.xmlelement.xmlelement import *


class TestXSDFalse(TestCase):
    def test_xml_element_with_xsd_false_simple(self):
        xml_score_part = XMLScorePart()
        with self.assertRaises(XMLElementChildrenRequired):
            xml_score_part.to_string()

        xml_score_part.xsd_check = False
        assert xml_score_part.to_string() == '<score-part />\n'

    def test_xml_part_list_xsd_false(self):
        xml_part_list = XMLPartList()
        with self.assertRaises(XMLElementChildrenRequired):
            xml_part_list.to_string()
        xml_part_list = XMLPartList(xsd_check=False)
        # xml_part_list.xsd_check = False
        assert xml_part_list.to_string() == '<part-list />\n'

        xml_part_list.add_child(XMLScorePart(id='p-1'))

        pg = xml_part_list.add_child(XMLPartGroup(number='1', type='start'))
        pg.add_child(XMLGroupSymbol('square'))
        pg.add_child(XMLGroupBarline('yes'))
        xml_part_list.add_child(XMLScorePart(id='p-2'))

        pg = xml_part_list.add_child(XMLPartGroup(number='2', type='start'))
        pg.add_child(XMLGroupSymbol('bracket'))
        pg.add_child(XMLGroupBarline('yes'))
        xml_part_list.add_child(XMLScorePart(id='p-3'))

        xml_part_list.add_child(XMLPartGroup(number='1', type='stop'))
        xml_part_list.add_child(XMLPartGroup(number='2', type='stop'))
        xml_part_list.add_child(XMLScorePart(id='p-4'))

        xml_part_list.add_child(XMLScorePart(id='p-5'))

        expected = """<part-list>
  <score-part id="p-1" />
  <part-group number="1" type="start">
    <group-symbol>square</group-symbol>
    <group-barline>yes</group-barline>
  </part-group>
  <score-part id="p-2" />
  <part-group number="2" type="start">
    <group-symbol>bracket</group-symbol>
    <group-barline>yes</group-barline>
  </part-group>
  <score-part id="p-3" />
  <part-group number="1" type="stop" />
  <part-group number="2" type="stop" />
  <score-part id="p-4" />
  <score-part id="p-5" />
</part-list>
"""

        assert xml_part_list.to_string() == expected

    def test_xml_part_list_init_xsd_false(self):
        xml_part_list = XMLPartList(xsd_check=False)
        assert xml_part_list.to_string() == '<part-list />\n'

    def test_xml_part_xsd_false_with_parent_xsd_true(self):
        xml_score = XMLScorePartwise()
        xml_part = xml_score.add_child(XMLPart(id='p1'))
        xml_part.add_child(XMLMeasure(number='1'))
        xml_part_list = XMLPartList(xsd_check=False)
        xml_score.add_child(xml_part_list)
        assert xml_part_list.to_string() == '<part-list />\n'
        print(xml_score.to_string())



import io
import io
import re
import xml.etree.ElementTree as ET
from contextlib import redirect_stdout
from typing import Optional

from musicxml.generate_classes.utils import musicxml_xsd_et_root, xml_xsd_et_root
from musicxml.util.core import cap_first, convert_to_xsd_class_name
from musicxml.util.helprervariables import xml_name_first_character_without_colon, name_character_without_colon, \
    name_character, \
    xml_name_first_character
from verysimpletree.tree import Tree

"""
XSD = XML Schema Definition
"""


class XSDTree(Tree):
    """
    XSDTree gets a xml.etree.ElementTree.Element by initiation as its xml_element_tree_element property and
    prepares all needed information for generating a :obj:`~musicxml.xsd.xsdtree.XSDTreeElement` class (:obj:`~musicxml.xsd.xsdtree.XSDTreeElement` can be :obj:`~musicxml.xsd.xsdsimpletype.XSDSimpleType`, :obj:`~musicxml.xsd.xsdcomplextype.XSDComplexType`, :obj:`~musicxml.xsd.xsdindicator.XSDGroup`,
    XMLAttribute and XMLAttributeGroup)
    """

    def __init__(self, xml_element_tree_element, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._namespace = None
        self._tag = None
        self._xsd_element_tree_element = None
        self._xml_tree_class_name = None
        self._xsd_indicator = None
        self._attributes = None
        self._text = None
        self._type = 'notset'
        self._name = 'notset'
        self.xml_element_tree_element = xml_element_tree_element

    # ------------------
    # private properties

    # ------------------
    # private methods

    def _get_xsd_tree_class_name(self):
        tag = cap_first(self.tag)

        name = 'XSD' + f'{tag}'
        if self.name:
            name += ''.join([cap_first(partial) for partial in self.name.split('-')])
        elif self.get_attributes().get('ref'):
            name += ''.join([cap_first(partial) for partial in self.get_attributes().get('ref').split('-')])
        else:
            raise AttributeError
        return name

    def _populate_children(self):
        for child in [XSDTree(node) for node in self.xml_element_tree_element.findall('./')]:
            self.add_child(child)

    def _check_child_to_be_added(self, child):
        if not isinstance(child, XSDTree):
            raise TypeError

    # ------------------
    # public properties
    @property
    def xsd_tree_base_class_names(self):

        if self.is_simple_type:
            # simple type
            if self.get_restriction():
                base = self.get_restriction().get_attributes()['base']
                return [convert_to_xsd_class_name(base)]
            elif self.get_union_member_types():
                return []
            else:
                raise AttributeError(
                    f"Simple type {self} has no restriction with base attribute or union with memberTypes.")
        elif self.is_complex_type:
            # complex type
            if self.get_simple_content_extension():
                base = self.get_simple_content_extension().get_attributes()['base']
                return [convert_to_xsd_class_name(base, type_='simple_type')]
            else:
                return []
                # raise AttributeError(f"Complex type {self} has no simple content extension with base attribute.")
        else:
            raise NotImplementedError

    @property
    def compact_repr(self):
        attrs = self.get_attributes()
        return f"{self.tag}{''.join([f'@{attribute}={attrs[attribute]}' for attribute in attrs])}"

    @property
    def is_simple_type(self):
        if self.tag == 'simpleType':
            return True
        return False

    @property
    def is_complex_type(self):
        if self.tag == 'complexType':
            return True
        return False

    @property
    def name(self):
        if self._name == 'notset':
            try:
                self._name = self.xml_element_tree_element.attrib['name']
            except KeyError:
                self._name = None
        return self._name

    @property
    def namespace(self):
        if not self._namespace:
            self._namespace = re.match(r'({.*})(.*)', self.xml_element_tree_element.tag).group(1)
        return self._namespace

    @property
    def tag(self):
        if not self._tag:
            self._tag = re.match(r'({.*})(.*)', self.xml_element_tree_element.tag).group(2)
        return self._tag

    @property
    def text(self):
        if not self._text:
            self._text = self.xml_element_tree_element.text
        return self._text

    @property
    def type(self):
        if self._type == 'notset':
            try:
                self._type = self.xml_element_tree_element.attrib['type']
            except KeyError:
                self._type = None
        return self._type

    @property
    def xml_element_tree_element(self):
        return self._xsd_element_tree_element

    @xml_element_tree_element.setter
    def xml_element_tree_element(self, value):
        if not isinstance(value, ET.Element):
            raise TypeError(
                f"XSDTree must be initiated with an xml_element_tree_element of type xml.etree.ElementTree.Element not "
                f"{type(value)}")
        self._xsd_element_tree_element = value

    @property
    def xsd_element_class_name(self):
        if self._xml_tree_class_name is None:
            self._xml_tree_class_name = self._get_xsd_tree_class_name()
        return self._xml_tree_class_name

    # ------------------
    # public methods
    def get_attributes(self):
        if self._attributes is None:
            self._attributes = self.xml_element_tree_element.attrib
            # if self.tag == 'element':
            #     if 'minOccurs' in self._attributes:
            #         self._attributes.pop('minOccurs')
            #     if 'maxOccurs' in self._attributes:
            #         self._attributes.pop('maxOccurs')
        return self._attributes

    def get_children(self):
        if not self._children:
            self._populate_children()
        return self._children

    def get_complex_content(self):
        for node in self.get_children():
            if node.tag == 'complexContent':
                return node

    def get_complex_content_extension(self):
        if self.get_complex_content().get_children()[0].tag == 'extension':
            return self.get_complex_content().get_children()[0]

    def get_doc(self):
        output = ''
        for node in self.traverse():
            if node.tag == 'documentation':
                output = node.text.strip()
                output.replace('\t', '    ')
                break
        permitted = self.get_permitted()
        pattern = self.get_pattern()
        if permitted:
            output += '\n    '
            output += '\n    '
            permitted = [f"``'{perm}'``" if perm else "``''``" for perm in permitted]
            output += f"Permitted Values: {', '.join(perm for perm in permitted)}\n"
        if pattern:
            output += '\n    '
            output += '\n    '
            output += f"    \nPattern: {pattern}\n"

        return output

    def get_restriction(self):
        for node in self.get_children():
            if node.tag == 'restriction':
                return node

    def get_pattern(self, parent_xsd_tree=None):
        def get_xsd_pattern(restriction_):
            if restriction_ and restriction_.get_children() and restriction_.get_children()[0].tag == 'pattern':
                return rf"{restriction.get_children()[0].get_attributes()['value']}"
            else:
                if parent_xsd_tree:
                    parent_restriction = parent_xsd_tree.get_restriction()
                    if parent_restriction and parent_restriction.get_children() and parent_restriction.get_children()[
                        0].tag == 'pattern':
                        return rf"{parent_restriction.get_children()[0].get_attributes()['value']}"

        def translate_pattern(pattern_):
            if pattern_ == "[\\i-[:]][\\c-[:]]*":
                return rf"{xml_name_first_character_without_colon}{name_character_without_colon}*"
            pattern_ = pattern_.replace('\\c', name_character)
            pattern_ = pattern_.replace('\\i', xml_name_first_character)
            return pattern_

        restriction = self.get_restriction()
        pattern = get_xsd_pattern(restriction)
        if pattern:
            return translate_pattern(pattern)
        else:
            return None

    def get_permitted(self):
        restriction = self.get_restriction()
        if restriction:
            enumerations = [child for child in restriction.get_children() if
                            child.tag == 'enumeration']
            return [enumeration.get_attributes()['value'] for enumeration in enumerations]

    def get_simple_content(self):
        for node in self.get_children():
            if node.tag == 'simpleContent':
                return node

    def get_simple_content_extension(self):
        for node in self.get_children():
            if node.tag == 'simpleContent':
                if node.get_children()[0].tag == 'extension':
                    return node.get_children()[0]

    def get_union(self):
        for node in self.get_children():
            if node.tag == 'union':
                return node

    def get_union_member_types(self):
        if self.get_union():
            return self.get_union().get_attributes()['memberTypes'].split(' ')

    def get_xsd(self):
        with io.StringIO() as buf, redirect_stdout(buf):
            ET.dump(self.xml_element_tree_element)
            output = buf.getvalue()
        output = output.strip()
        output += '\n'
        return output

    def get_xsd_indicator(self):
        return self._xsd_indicator

    # ------------------
    # magic methods
    def __deepcopy__(self, copy_parent=False):
        def copy_et_element(el):
            output = ET.Element(el.tag, el.attrib)
            output.text = el.text
            return output

        copied = self.__class__(xml_element_tree_element=copy_et_element(self.xml_element_tree_element))
        copied._tag = self.tag
        if copy_parent and self.get_parent():
            copied._parent = self.get_parent().__deepcopy__(copy_parent=True)
        for ch in self.get_children():
            copied.add_child(ch.__deepcopy__())
        return copied

    def __repr__(self):
        attrs = self.get_attributes()
        output = f"{self.__class__.__name__}(tag={self.tag}"
        if attrs:
            output += f", {' '.join([f'{attribute}={attrs[attribute]}' for attribute in attrs])})"
        else:
            output += ')'
        return output

    def __str__(self):
        return f"{self.__class__.__name__} {self.compact_repr}"


class XSDTreeElement:
    """
    Abstract class of all generated XSD Classes
    """
    XSD_TREE: Optional[XSDTree] = None
    _SEARCH_FOR_ELEMENT = ""
    _XSD_TREE = None

    @classmethod
    def get_xsd_tree(cls):
        if cls._XSD_TREE:
            return cls._XSD_TREE

        found_xsd_element = musicxml_xsd_et_root.find(cls._SEARCH_FOR_ELEMENT)
        if found_xsd_element is None:
            return cls.XSD_TREE
        else:
            cls._XSD_TREE = XSDTree(found_xsd_element)
            return cls._XSD_TREE

    @classmethod
    def get_xsd(cls):
        return cls.get_xsd_tree().get_xsd()


extra_elements = {
    'score-partwise':
        {'search_for': ".//{*}element[@name='score-partwise']",
         },
    'part':
        {'search_for': ".//{*}element[@name='score-partwise']//{*}element[@name='part']",
         },
    'measure':
        {'search_for': ".//{*}element[@name='score-partwise']//{*}element[@name='measure']",
         },
    'directive':
        {'search_for': ".//{*}complexType[@name='attributes']//{*}element[@name='directive']",
         }
}


def _generate_xsd_tree():
    """
    Makes a dictionary out of musicxml_xsd_et_root children with appropriate keys
    """
    output = {'simpleType': {}, 'complexType': {}, 'element': {}, 'group': {}, 'attribute': {}, 'attributeGroup': {}}
    for root in [xml_xsd_et_root, musicxml_xsd_et_root]:
        for node in root.iter():
            tag_ = node.tag.split('}')[1]
            name = node.attrib.get('name')
            if name and tag_ in output:
                ET.indent(node, space='    ')
                output[tag_][name] = XSDTree(xml_element_tree_element=node)
    for el_name in extra_elements:
        tag_ = 'element'
        name = el_name
        node = musicxml_xsd_et_root.find(extra_elements[el_name]['search_for'])
        ET.indent(node, space='    ')
        output[tag_][name] = XSDTree(xml_element_tree_element=node)

    return output


XSD_TREE_DICT = _generate_xsd_tree()



from musicxml.util.core import cap_first, convert_to_xml_class_name
from musicxml.xsd.xsdtree import XSDTree, XSDTreeElement, XSD_TREE_DICT
import xml.etree.ElementTree as ET


class XSDSequence:
    def __init__(self, xsd_tree):
        self._xsd_tree = None
        self._elements = None
        self.xsd_tree = xsd_tree

    @property
    def elements(self):
        if not self._elements:
            self._elements = []
            for child in self.xsd_tree.get_children():
                if child.tag == 'element':
                    element = convert_to_xml_class_name(child.name)
                    min_occurrence = child.get_attributes().get('minOccurs')
                    if min_occurrence is None: min_occurrence = '1'
                    max_occurrence = child.get_attributes().get('maxOccurs')
                    if max_occurrence is None: max_occurrence = '1'
                    self._elements.append((element, min_occurrence, max_occurrence))

                elif child.tag == 'group':
                    xsd_group_name = 'XSDGroup' + ''.join([cap_first(partial) for partial in child.get_attributes()['ref'].split('-')])
                    elements = eval(xsd_group_name)().sequence.elements
                    min_occurrence = child.get_attributes().get('minOccurs')
                    max_occurrence = child.get_attributes().get('maxOccurs')
                    if min_occurrence is not None:
                        if len(elements) > 1:
                            raise NotImplementedError
                        list_el = list(elements[0])
                        list_el[1] = min_occurrence
                        elements[0] = tuple(list_el)
                    if max_occurrence is not None:
                        if len(elements) > 1:
                            raise NotImplementedError
                        list_el = list(elements[0])
                        list_el[2] = max_occurrence
                        elements[0] = tuple(list_el)
                    self._elements.extend(elements)
                else:
                    raise NotImplementedError(child.tag)
        return self._elements

    @property
    def xsd_tree(self):
        return self._xsd_tree

    @xsd_tree.setter
    def xsd_tree(self, value):
        if not isinstance(value, XSDTree):
            raise TypeError
        if value.tag != 'sequence':
            raise ValueError
        self._xsd_tree = value

    def __copy__(self):
        return self.__class__(xsd_tree=self.xsd_tree)


class XSDChoice:
    def __init__(self, xsd_tree):
        self._xsd_tree = None
        self.xsd_tree = xsd_tree

    @property
    def xsd_tree(self):
        return self._xsd_tree

    @xsd_tree.setter
    def xsd_tree(self, value):
        if not isinstance(value, XSDTree):
            raise TypeError
        if value.tag != 'choice':
            raise ValueError
        self._xsd_tree = value

    def __copy__(self):
        return self.__class__(xsd_tree=self.xsd_tree)


class XSDGroup(XSDTreeElement):

    def __init__(self):
        self._sequence = None

    @property
    def name(self):
        return self.XSD_TREE.name

    @property
    def sequence(self):
        if not self._sequence:
            for child in self.XSD_TREE.get_children():
                if child.tag == 'sequence':
                    self._sequence = XSDSequence(child)
        return self._sequence

    @property
    def xsd_tree(self):
        return self.XSD_TREE

    def __copy__(self):
        copied = self.__class__()
        copied._sequence = self.sequence
        copied.XSD_TREE = self.XSD_TREE
        return copied
# -----------------------------------------------------
# AUTOMATICALLY GENERATED WITH generate_indicators.py
# -----------------------------------------------------


class XSDGroupEditorial(XSDGroup):
    """The editorial group specifies editorial information for a musical element."""
    
    XSD_TREE = XSD_TREE_DICT['group']['editorial']


class XSDGroupEditorialVoice(XSDGroup):
    """The editorial-voice group supports the common combination of editorial and voice information for a musical element."""
    
    XSD_TREE = XSD_TREE_DICT['group']['editorial-voice']


class XSDGroupEditorialVoiceDirection(XSDGroup):
    """The editorial-voice-direction group supports the common combination of editorial and voice information for a direction element. It is separate from the editorial-voice element because extensions and restrictions might be different for directions than for the note and forward elements."""
    
    XSD_TREE = XSD_TREE_DICT['group']['editorial-voice-direction']


class XSDGroupFootnote(XSDGroup):
    """The footnote element specifies editorial information that appears in footnotes in the printed score. It is defined within a group due to its multiple uses within the MusicXML schema."""
    
    XSD_TREE = XSD_TREE_DICT['group']['footnote']


class XSDGroupLevel(XSDGroup):
    """The level element specifies editorial information for different MusicXML elements. It is defined within a group due to its multiple uses within the MusicXML schema."""
    
    XSD_TREE = XSD_TREE_DICT['group']['level']


class XSDGroupStaff(XSDGroup):
    """The staff element is defined within a group due to its use by both notes and direction elements."""
    
    XSD_TREE = XSD_TREE_DICT['group']['staff']


class XSDGroupTuning(XSDGroup):
    """The tuning group contains the sequence of elements common to the staff-tuning and accord elements."""
    
    XSD_TREE = XSD_TREE_DICT['group']['tuning']


class XSDGroupVirtualInstrumentData(XSDGroup):
    """Virtual instrument data can be part of either the score-instrument element at the start of a part, or an instrument-change element within a part."""
    
    XSD_TREE = XSD_TREE_DICT['group']['virtual-instrument-data']


class XSDGroupVoice(XSDGroup):
    """A voice is a sequence of musical events (e.g. notes, chords, rests) that proceeds linearly in time. The voice element is used to distinguish between multiple voices in individual parts. It is defined within a group due to its multiple uses within the MusicXML schema."""
    
    XSD_TREE = XSD_TREE_DICT['group']['voice']


class XSDGroupClef(XSDGroup):
    """Clefs are represented by a combination of sign, line, and clef-octave-change elements."""
    
    XSD_TREE = XSD_TREE_DICT['group']['clef']


class XSDGroupNonTraditionalKey(XSDGroup):
    """The non-traditional-key group represents a single alteration within a non-traditional key signature. A sequence of these groups makes up a non-traditional key signature"""
    
    XSD_TREE = XSD_TREE_DICT['group']['non-traditional-key']


class XSDGroupSlash(XSDGroup):
    """The slash group combines elements used for more complete specification of the slash and beat-repeat measure-style elements. They have the same values as the type and dot elements, and define what the beat is for the display of repetition marks. If not present, the beat is based on the current time signature."""
    
    XSD_TREE = XSD_TREE_DICT['group']['slash']


class XSDGroupTimeSignature(XSDGroup):
    """Time signatures are represented by the beats element for the numerator and the beat-type element for the denominator."""
    
    XSD_TREE = XSD_TREE_DICT['group']['time-signature']


class XSDGroupTraditionalKey(XSDGroup):
    """The traditional-key group represents a traditional key signature using the cycle of fifths."""
    
    XSD_TREE = XSD_TREE_DICT['group']['traditional-key']


class XSDGroupTranspose(XSDGroup):
    """The transpose group represents what must be added to a written pitch to get a correct sounding pitch."""
    
    XSD_TREE = XSD_TREE_DICT['group']['transpose']


class XSDGroupBeatUnit(XSDGroup):
    """The beat-unit group combines elements used repeatedly in the metronome element to specify a note within a metronome mark."""
    
    XSD_TREE = XSD_TREE_DICT['group']['beat-unit']


class XSDGroupHarmonyChord(XSDGroup):
    """A harmony element can contain many stacked chords (e.g. V of II). A sequence of harmony-chord groups is used for this type of secondary function, where V of II would be represented by a harmony-chord with a 5 numeral followed by a harmony-chord with a 2 numeral.

A root is a pitch name like C, D, E, while a numeral is a scale degree like 1, 2, 3. The root element is generally used with pop chord symbols, while the numeral element is generally used with classical functional harmony and Nashville numbers. It is an either/or choice to avoid data inconsistency. The function element, which represents Roman numerals with roman numeral text, has been deprecated as of MusicXML 4.0."""
    
    XSD_TREE = XSD_TREE_DICT['group']['harmony-chord']


class XSDGroupAllMargins(XSDGroup):
    """The all-margins group specifies both horizontal and vertical margins in tenths."""
    
    XSD_TREE = XSD_TREE_DICT['group']['all-margins']


class XSDGroupLayout(XSDGroup):
    """The layout group specifies the sequence of page, system, and staff layout elements that is common to both the defaults and print elements."""
    
    XSD_TREE = XSD_TREE_DICT['group']['layout']


class XSDGroupLeftRightMargins(XSDGroup):
    """The left-right-margins group specifies horizontal margins in tenths."""
    
    XSD_TREE = XSD_TREE_DICT['group']['left-right-margins']


class XSDGroupDuration(XSDGroup):
    """The duration element is defined within a group due to its uses within the note, figured-bass, backup, and forward elements."""
    
    XSD_TREE = XSD_TREE_DICT['group']['duration']


class XSDGroupDisplayStepOctave(XSDGroup):
    """The display-step-octave group contains the sequence of elements used by both the rest and unpitched elements. This group is used to place rests and unpitched elements on the staff without implying that these elements have pitch. Positioning follows the current clef. If percussion clef is used, the display-step and display-octave elements are interpreted as if in treble clef, with a G in octave 4 on line 2."""
    
    XSD_TREE = XSD_TREE_DICT['group']['display-step-octave']


class XSDGroupFullNote(XSDGroup):
    """The full-note group is a sequence of the common note elements between cue/grace notes and regular (full) notes: pitch, chord, and rest information, but not duration (cue and grace notes do not have duration encoded). Unpitched elements are used for unpitched percussion, speaking voice, and other musical elements lacking determinate pitch."""
    
    XSD_TREE = XSD_TREE_DICT['group']['full-note']


class XSDGroupMusicData(XSDGroup):
    """The music-data group contains the basic musical data that is either associated with a part or a measure, depending on whether the partwise or timewise hierarchy is used."""
    
    XSD_TREE = XSD_TREE_DICT['group']['music-data']


class XSDGroupPartGroup(XSDGroup):
    """The part-group element is defined within a group due to its multiple uses within the part-list element."""
    
    XSD_TREE = XSD_TREE_DICT['group']['part-group']


class XSDGroupScoreHeader(XSDGroup):
    """The score-header group contains basic score metadata about the work and movement, score-wide defaults for layout and fonts, credits that appear on the first or following pages, and the part list."""
    
    XSD_TREE = XSD_TREE_DICT['group']['score-header']


class XSDGroupScorePart(XSDGroup):
    """The score-part element is defined within a group due to its multiple uses within the part-list element."""
    
    XSD_TREE = XSD_TREE_DICT['group']['score-part']

__all__=['XSDSequence', 'XSDChoice', 'XSDGroup', 'XSDGroupEditorial', 'XSDGroupEditorialVoice', 'XSDGroupEditorialVoiceDirection', 'XSDGroupFootnote', 'XSDGroupLevel', 'XSDGroupStaff', 'XSDGroupTuning', 'XSDGroupVirtualInstrumentData', 'XSDGroupVoice', 'XSDGroupClef', 'XSDGroupNonTraditionalKey', 'XSDGroupSlash', 'XSDGroupTimeSignature', 'XSDGroupTraditionalKey', 'XSDGroupTranspose', 'XSDGroupBeatUnit', 'XSDGroupHarmonyChord', 'XSDGroupAllMargins', 'XSDGroupLayout', 'XSDGroupLeftRightMargins', 'XSDGroupDuration', 'XSDGroupDisplayStepOctave', 'XSDGroupFullNote', 'XSDGroupMusicData', 'XSDGroupPartGroup', 'XSDGroupScoreHeader', 'XSDGroupScorePart']



class XSDSimpleTypeFontSize(XSDSimpleType):
    """The font-size can be one of the CSS font sizes (xx-small, x-small, small, medium, large, x-large, xx-large) or a numeric point size.

    .. todo::
       Better documentation.
    """
    _UNION = [XSDSimpleTypeCssFontSize, XSDSimpleTypeDecimal]
    XSD_TREE = XSDTree(ET.fromstring("""
<xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="font-size">
    <xs:annotation>
        <xs:documentation>The font-size can be one of the CSS font sizes (xx-small, x-small, small, medium, large, x-large, xx-large) or a numeric point size.</xs:documentation>
    </xs:annotation>
    <xs:union memberTypes="xs:decimal css-font-size" />
</xs:simpleType>
"""
                                     ))


class XSDSimpleTypeYesNoNumber(XSDSimpleType):
    """The yes-no-number type is used for attributes that can be either boolean or numeric values.

    .. todo::
       Better documentation.
    """
    _UNION = [XSDSimpleTypeYesNo, XSDSimpleTypeDecimal]
    XSD_TREE = XSDTree(ET.fromstring("""
<xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="yes-no-number">
    <xs:annotation>
        <xs:documentation>The yes-no-number type is used for attributes that can be either boolean or numeric values.</xs:documentation>
    </xs:annotation>
    <xs:union memberTypes="yes-no xs:decimal" />
</xs:simpleType>
"""
                                     ))


class XSDSimpleTypePositiveIntegerOrEmpty(XSDSimpleTypePositiveInteger):
    """The positive-integer-or-empty values can be either a positive integer or an empty string.

    .. todo::
       Better documentation.
    """
    _FORCED_PERMITTED = ['']
    XSD_TREE = XSDTree(ET.fromstring("""
<xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="positive-integer-or-empty">
    <xs:annotation>
        <xs:documentation>The positive-integer-or-empty values can be either a positive integer or an empty string.</xs:documentation>
    </xs:annotation>
    <xs:union memberTypes="xs:positiveInteger">
        <xs:simpleType>
            <xs:restriction base="xs:string">
                <xs:enumeration value="" />
            </xs:restriction>
        </xs:simpleType>
    </xs:union>
</xs:simpleType>
"""
                                     ))

    def __init__(self, value='', *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


class XSDSimpleTypeNumberOrNormal(XSDSimpleTypeDecimal):
    """The number-or-normal values can be either a decimal number or the string "normal". This is used by the line-height and letter-spacing attributes.

    .. todo::
       Better documentation.
    """
    _FORCED_PERMITTED = ['normal']
    XSD_TREE = XSDTree(ET.fromstring("""
<xs:simpleType xmlns:xs="http://www.w3.org/2001/XMLSchema" name="number-or-normal">
    <xs:annotation>
        <xs:documentation>The number-or-normal values can be either a decimal number or the string "normal". This is used by the line-height and letter-spacing attributes.</xs:documentation>
    </xs:annotation>
    <xs:union memberTypes="xs:decimal">
        <xs:simpleType>
            <xs:restriction base="xs:token">
                <xs:enumeration value="normal" />
            </xs:restriction>
        </xs:simpleType>
    </xs:union>
</xs:simpleType>
"""
                                     ))



import xml.etree.ElementTree as ET

from musicxml.util.core import convert_to_xml_class_name, cap_first
from musicxml.xmlelement.exceptions import XMLChildContainerFactoryError, XMLChildContainerWrongElementError, \
    XMLChildContainerChoiceHasAnotherChosenChild, XMLChildContainerMaxOccursError
from musicxml.xsd.xsdelement import XSDElement
from musicxml.xsd.xsdindicator import *
from musicxml.xsd.xsdtree import XSDTree
from verysimpletree.tree import Tree



def _convert_xsd_child_to_xsd_container(xsd_child):
    min_occurrences = xsd_child.get_attributes().get('minOccurs')
    max_occurrences = xsd_child.get_attributes().get('maxOccurs')

    copied_xsd_child = xsd_child.__deepcopy__()
    if min_occurrences is not None:
        copied_xsd_child.get_attributes().pop('minOccurs')
    if max_occurrences is not None:
        copied_xsd_child.get_attributes().pop('maxOccurs')
    if xsd_child.tag == 'element':
        return XMLChildContainer(content=XSDElement(copied_xsd_child), min_occurrences=min_occurrences,
                                 max_occurrences=max_occurrences)
    elif xsd_child.tag == 'sequence':
        return XMLChildContainer(content=XSDSequence(copied_xsd_child), min_occurrences=min_occurrences,
                                 max_occurrences=max_occurrences)
    elif xsd_child.tag == 'choice':
        return XMLChildContainer(content=XSDChoice(copied_xsd_child), min_occurrences=min_occurrences,
                                 max_occurrences=max_occurrences)
    elif xsd_child.tag == 'group':
        xsd_group_name = 'XSDGroup' + ''.join(
            [cap_first(partial) for partial in xsd_child.get_attributes()['ref'].split('-')])
        return XMLChildContainer(content=eval(xsd_group_name)(), min_occurrences=min_occurrences,
                                 max_occurrences=max_occurrences)
    else:
        raise NotImplementedError(xsd_child.tag)


def _check_if_container_requires_elements(xsd_container):
    if isinstance(xsd_container.content, XSDSequence):
        return _check_if_sequence_requires_elements(xsd_container)
    elif isinstance(xsd_container.content, XSDGroup):
        return _check_if_group_requires_elements(xsd_container)
    elif isinstance(xsd_container.content, XSDChoice):
        return _check_if_choice_requires_elements(xsd_container)
    else:
        raise NotImplementedError(xsd_container)


def _check_if_choice_requires_elements(xsd_container_choice):
    element_chosen = False
    for child in xsd_container_choice.get_children():
        if isinstance(child.content, XSDGroup):
            if child.get_children()[0].force_validate:
                _check_if_container_requires_elements(child.get_children()[0])
        elif child.force_validate:
            _check_if_container_requires_elements(child)
        else:
            if child.min_occurrences == 0:
                pass
            elif int(child.min_occurrences) == 1:
                if isinstance(child.content, XSDElement):
                    if len(child.content.xml_elements) == 0:
                        pass
                    elif len(child.content.xml_elements) == 1:
                        element_chosen = True
                    else:
                        raise NotImplementedError(child)
                else:
                    _check_if_container_requires_elements(child)
            else:
                raise NotImplementedError(f'child {child} with min_occurrence greater than 1')

    if element_chosen:
        xsd_container_choice.requirements_fulfilled = True


def _check_if_group_requires_elements(xsd_group_container):
    if xsd_group_container.min_occurrences == 0 and not xsd_group_container.get_children()[0].force_validate:
        return
    return _check_if_sequence_requires_elements(xsd_group_container.get_children()[0])


def _check_if_sequence_requires_elements(xsd_sequence_container):
    if xsd_sequence_container.force_validate:
        for child in xsd_sequence_container.get_children():
            if isinstance(child.content, XSDElement):
                if len(child.content.xml_elements) < child.min_occurrences:
                    child.requirements_fulfilled = False
                else:
                    pass
            else:
                _check_if_container_requires_elements(child)

    def validate_child(ch):
        if isinstance(ch.content, XSDElement):
            if child.choices_in_reversed_path:
                pass
            elif len(ch.content.xml_elements) < ch.min_occurrences:
                ch.requirements_fulfilled = False
            else:
                ch.requirements_fulfilled = True
        else:
            _check_if_container_requires_elements(ch)

    if xsd_sequence_container.min_occurrences > 0:
        for child in xsd_sequence_container.get_children():
            if child.force_validate is True:
                _check_if_container_requires_elements(child)
            elif child.min_occurrences == 0:
                pass
            elif child.min_occurrences == 1:
                validate_child(child)
            else:
                raise NotImplementedError(f'child {child} with min_occurrence greater than 1')


class DuplicationXSDSequence(XSDSequence):
    sequence_xsd = """
            <xs:sequence xmlns:xs="http://www.w3.org/2001/XMLSchema">
            </xs:sequence>
    """

    def __init__(self):
        xsd_tree_ = XSDTree(ET.fromstring(self.sequence_xsd))
        super().__init__(xsd_tree_)


class XMLChildContainer(Tree):
    def __init__(self, content, min_occurrences=None, max_occurrences=None, populate_children=True, *args, **kwargs):
        super().__init__(content=content, *args, **kwargs)
        # self._content = None
        self._chosen_child = None
        self._required_element_names = None
        self._requirements_fulfilled = None
        self.min_occurrences = 1 if min_occurrences is None else int(min_occurrences)
        self.max_occurrences = 1 if max_occurrences is None else 'unbounded' if max_occurrences == 'unbounded' else int(
            max_occurrences)
        # self.content = content
        self._force_validate = None
        self._parent_xml_element = None
        if populate_children:
            self._populate_children()

    # private methods
    def _add_duplication_parent(self):

        if not self.get_parent():
            parent_container = XMLChildContainer(content=DuplicationXSDSequence())
            parent_container.add_child(self)
        elif not (isinstance(self.get_parent().content, DuplicationXSDSequence)):
            parent_container = XMLChildContainer(content=DuplicationXSDSequence())
            index = self.get_parent().get_children().index(self)
            parent = self.get_parent()
            parent.remove(self)
            parent.get_children().insert(index, parent_container)
            parent_container._parent = parent
            parent_container.add_child(self)
            if isinstance(parent.content, XSDChoice) and parent.chosen_child == self:
                parent.chosen_child = parent_container
        else:
            pass

    @staticmethod
    def _check_content_type(val):
        types = [XSDSequence, XSDChoice, XSDElement, XSDGroup]
        for type_ in types:
            if isinstance(val, type_):
                return
        raise TypeError(val)

    def _check_child_to_be_added(self, child):
        if not isinstance(child, XMLChildContainer):
            raise TypeError

    def _check_choices_intelligently(self, xml_element=None):
        """
        Check if existing xml elements can be attached to other choice paths in order to fulfill all requirements. Only possible leaves
        forwards will be checked. If xml_element is given it will be treated as a new element which is going to be attached to the
        container.
        """
        if self.get_parent_xml_element():
            print(f'_check_choices_intelligently: intelligent choice for {self.get_parent_xml_element()}')

        def get_same_name_next_leaves(leaf_):
            output = []
            index = -1
            record = False
            for l in self.iterate_leaves():
                if l.content.name == leaf_.content.name:
                    index += 1
                    if not record:
                        record = True
                    else:
                        output.append((l, index))
            return output

        def get_sorted_xml_elements(sorted_options_):
            unsorted_elements = [el for el in self.get_attached_elements() if el.name != sorted_options_[-1][0]]
            output = []
            for option in sorted_options_:
                selected_elements = [el for el in unsorted_elements if el.name == option[1]]
                for el in selected_elements:
                    unsorted_elements.remove(el)
                    output.append(el)
            output.extend(unsorted_elements)
            return output

        current_leaves_with_xml_elements = [leaf for leaf in self.iterate_leaves() if leaf.content.xml_elements]
        optional_next_leaves = [get_same_name_next_leaves(leaf) for leaf in current_leaves_with_xml_elements if
                                get_same_name_next_leaves(leaf)]
        if optional_next_leaves:
            indices = {list_of_leaves[0][0].content.name: [l[1] for l in list_of_leaves] for list_of_leaves in
                       optional_next_leaves}
            sorted_options = sorted([(k, v) for k, v in indices.items()], key=lambda v: -len(v))
            sorted_xml_elements = get_sorted_xml_elements(sorted_options)
            efficient_xml_element_name, forward_indices = sorted_options.pop()
            efficient_xml_elements = [el for el in self.get_attached_elements() if
                                      el.name == efficient_xml_element_name]
            for element in efficient_xml_elements:
                for forward_index in forward_indices:
                    copied_container = self._create_empty_copy()
                    copied_container.add_element(element, forward_index)
                    try:
                        for el in sorted_xml_elements:
                            copied_container.add_element(el, intelligent_choice=False)
                        if xml_element:
                            copied_container.add_element(xml_element, intelligent_choice=False)
                            return copied_container
                        if not copied_container.check_required_elements():
                            return copied_container
                    except XMLChildContainerChoiceHasAnotherChosenChild:
                        pass
        return None

    def _create_empty_copy(self):
        """
        Creates a copy without attached elements or duplicated nodes
        :return: XMLChildContainer
        """
        if isinstance(self.content, XSDChoice) or isinstance(self.content, XSDSequence):
            copied_content = self.content.__class__(self.content.xsd_tree)
        else:
            copied_content = eval(self.content.__class__.__name__)()
        return XMLChildContainer(copied_content, self.min_occurrences, self.max_occurrences)

    def _duplicate_parent_in_path(self):
        for node in list(self.get_reversed_path_to_root())[:-1]:
            if node.get_parent().max_occurrences == 'unbounded':
                return node.get_parent().duplicate()
        return None

    def _update_requirements_in_path(self):
        if not isinstance(self.content, XSDElement):
            raise ValueError
        if self.max_is_reached:
            self.requirements_fulfilled = True
        if self.content.xml_elements:
            for node in self.get_reversed_path_to_root():
                if node.get_parent():
                    if isinstance(node.get_parent().content, XSDChoice):
                        if node.get_parent().chosen_child:
                            if node.get_parent().chosen_child != node:
                                raise XMLChildContainerChoiceHasAnotherChosenChild
                            else:
                                break
                        else:
                            node.get_parent().chosen_child = node
                            if node.get_parent().requirements_fulfilled is False:
                                node.get_parent().requirements_fulfilled = True
                                break
                            node.get_parent().requirements_fulfilled = True
                    elif isinstance(node.get_parent().content, XSDSequence):
                        if node.get_parent().force_validate:
                            break
                        else:
                            node.get_parent().set_force_validate(node, True)

    def _populate_children(self):
        for xsd_child in [child for child in self.content.xsd_tree.get_children() if
                          child.tag != 'annotation' and child.tag != 'complexType']:
            container = _convert_xsd_child_to_xsd_container(xsd_child)
            self.add_child(container)

    def _set_requirements_fulfilled(self):
        for node in self.traverse():
            if isinstance(node.content, XSDChoice) and node.requirements_fulfilled is None and False not in [
                choice.requirements_fulfilled for choice in
                node.choices_in_reversed_path] and node.min_occurrences != 0:
                for leaf in node.iterate_leaves():
                    if leaf.content.xml_elements:
                        node._requirements_fulfilled = True
                if node._requirements_fulfilled is None:
                    for child in node.get_children():
                        if child.min_occurrences != 0:
                            node._requirements_fulfilled = False
                        break
                if node._requirements_fulfilled is None:
                    node._requirements_fulfilled = True
            else:
                node._requirements_fulfilled = True

    # public properties

    @property
    def compact_repr(self):
        """
        :return: A compact representation of :obj:`~content`.
        """
        def get_indentation() -> str:
            indentation = ''
            for i in range(self.get_level()):
                indentation += '    '
            return indentation

        if isinstance(self.content, XSDSequence):
            type_ = 'Sequence'
            return f"{type_}@minOccurs={self.min_occurrences}@maxOccurs={self.max_occurrences}"

        if isinstance(self.content, XSDChoice):
            type_ = 'Choice'
            output = f"{type_}@minOccurs={self.min_occurrences}@maxOccurs={self.max_occurrences}"
            if self.requirements_fulfilled is False:
                output += '\n'
                output += get_indentation() + '    '
                output += '!Required!'
            return output

        if isinstance(self.content, XSDGroup):
            type_ = 'Group'
            return f"{type_}@name={self.content.name}@minOccurs={self.min_occurrences}@maxOccurs={self.max_occurrences}"

        if isinstance(self.content, XSDElement):
            type_ = 'Element'
            output = f"{type_}@name={self.content.name}@minOccurs={self.min_occurrences}@maxOccurs={self.max_occurrences}"
            for xml_element in self.content.xml_elements:
                output += '\n'
                output += get_indentation() + '    '
                output += xml_element.__class__.__name__
            if self.requirements_fulfilled is False:
                output += '\n'
                output += get_indentation() + '    '
                output += '!Required!'
            return output

    @property
    def content(self):
        """
        :return: Content of a :obj:`~musicxml.xmlelement.xmlchildcontainer.XMLChildContainer` is its core property. It can be of types: :obj:`~musicxml.xsd.xsdindicator.XSDSequence`, :obj:`~musicxml.xsd.xsdindicator.XSDChoice`, :obj:`~musicxml.xsd.xsdindicator.XSDGroup` or :obj:`~musicxml.xsd.xsdelement.XSDElement` which are used to translate the behaviour of according xsd tags: ``sequence``, ``choice``, ``group`` and ``element``.
        """
        return self._content

    @Tree.content.setter
    def content(self, val):
        self._check_content_type(val)
        self._content = val
        self._content.parent_container = self

    @property
    def choices_in_reversed_path(self):
        return [node for node in list(self.get_reversed_path_to_root())[1:] if isinstance(node.content, XSDChoice)]

    @property
    def chosen_child(self):
        return self._chosen_child

    @chosen_child.setter
    def chosen_child(self, val):
        if not isinstance(self.content, XSDChoice):
            raise TypeError
        self._chosen_child = val

    @property
    def force_validate(self):
        return self._force_validate

    @property
    def max_is_reached(self):
        if not isinstance(self.content, XSDElement):
            raise TypeError
        if self.max_occurrences == 'unbounded':
            return False
        else:
            if len(self.content.xml_elements) == self.max_occurrences:
                return True
            elif len(self.content.xml_elements) > self.max_occurrences:
                raise ValueError
            else:
                return False

    @property
    def requirements_fulfilled(self):
        return self._requirements_fulfilled

    @requirements_fulfilled.setter
    def requirements_fulfilled(self, val: bool):
        if not isinstance(val, bool):
            raise TypeError
        self._requirements_fulfilled = val

    # public methods

    def add_element(self, xml_element, forward=None, intelligent_choice=True):
        # resets frozen tree iteration lists like traverse() iterate_leaves() etc.
        self._reset_iterators()

        if self._requirements_fulfilled is None:
            self.check_required_elements()

        def select_valid_leaves(leaves):
            output = []
            choice_with_chosen_child = None
            for index, leaf in enumerate(leaves):
                for n in leaf.get_reversed_path_to_root():
                    if n.get_parent() and isinstance(n.get_parent().content, XSDChoice) and n.get_parent().chosen_child:
                        choice_with_chosen_child = n.get_parent()
                        if n == choice_with_chosen_child.chosen_child:
                            output.append(leaf)
                        break

            if not choice_with_chosen_child:
                return leaves

            elif not output:
                if choice_with_chosen_child.max_occurrences == 1:
                    if choice_with_chosen_child.up and isinstance(choice_with_chosen_child.up.content, XSDSequence) and \
                            choice_with_chosen_child.up.max_occurrences == 'unbounded':
                        return []
                    else:
                        return None
                else:
                    return []
            else:
                return output

        if 'XMLElement' not in [cls.__name__ for cls in xml_element.__class__.__mro__]:
            raise TypeError(xml_element.__class__)

        same_name_leaves = [leaf for leaf in self.iterate_leaves() if leaf.content.name == xml_element.name]

        if not same_name_leaves:
            raise XMLChildContainerWrongElementError()

        selected_same_name_leaves = select_valid_leaves(same_name_leaves)
        if selected_same_name_leaves is None:
            intelligently_selected = None
            if forward is None and intelligent_choice is True:
                copy_with_intelligence = self._check_choices_intelligently(xml_element)
                if copy_with_intelligence:
                    for old_child, new_child in zip(self.get_children(), copy_with_intelligence.get_children()):
                        self.replace_child(old_child, new_child)
                    return \
                        [l for l in copy_with_intelligence.iterate_leaves() if xml_element in l.content.xml_elements][0]

            if not intelligently_selected:
                msg = f"{self} By adding {xml_element.__class__.__name__} to {self.get_parent_xml_element().__class__.__name__}" if \
                    self.get_parent_xml_element() else f"{self} By adding {xml_element.__class__.__name__}"
                raise XMLChildContainerChoiceHasAnotherChosenChild(msg)
            else:
                self.check_required_elements()
                return intelligently_selected

        if selected_same_name_leaves == []:
            duplicated_parent = same_name_leaves[-1]._duplicate_parent_in_path()
            if duplicated_parent:
                selected_same_name_leaves = [leaf for leaf in duplicated_parent.iterate_leaves() if
                                             leaf.content.name == xml_element.name and not
                                             leaf.max_is_reached]
                if self._parent_xml_element and self.up:
                    self._parent_xml_element._child_container_tree = self.up

            else:
                raise XMLChildContainerChoiceHasAnotherChosenChild

        if forward is not None:
            selected = same_name_leaves[forward]
            if selected not in selected_same_name_leaves:
                raise XMLChildContainerChoiceHasAnotherChosenChild('Wrong forwarding')
        else:
            selected_same_name_leaves_max_not_reached = [leaf for leaf in selected_same_name_leaves if
                                                         not leaf.max_is_reached]
            if not selected_same_name_leaves_max_not_reached:
                duplicated_parent = selected_same_name_leaves[-1]._duplicate_parent_in_path()
                if duplicated_parent:
                    selected_same_name_leaves_max_not_reached = [leaf for leaf in duplicated_parent.iterate_leaves() if
                                                                 leaf.content.name ==
                                                                 xml_element.name and not leaf.max_is_reached]
                    if self._parent_xml_element and self.up:
                        self._parent_xml_element._child_container_tree = self.up
                else:
                    raise XMLChildContainerMaxOccursError()

            selected = selected_same_name_leaves_max_not_reached[0]

        selected.content.add_xml_element(xml_element)
        selected._update_requirements_in_path()
        return selected

    def check_required_elements(self, intelligent_choice=False):
        if self._requirements_fulfilled is None:
            self._set_requirements_fulfilled()
        _check_if_container_requires_elements(self)
        requirements_exist = False
        for node in self.traverse():
            if node.requirements_fulfilled is False:
                requirements_exist = True
        if requirements_exist and intelligent_choice:
            if isinstance(self.content, XSDChoice):
                return requirements_exist
            copy_with_intelligence = self._check_choices_intelligently()
            if copy_with_intelligence:
                for old_child, new_child in zip(self.get_children(), copy_with_intelligence.get_children()):
                    self.replace_child(old_child, new_child)
                return False

        return requirements_exist

    check_requirements = check_required_elements

    def duplicate(self):
        if not isinstance(self.content, XSDSequence) and not isinstance(self.content, XSDChoice) and not isinstance(
                self.content, XSDGroup):
            raise TypeError(self.content)

        if self.max_occurrences != 'unbounded':
            raise ValueError

        self._add_duplication_parent()

        copied_self = self._create_empty_copy()
        copied_self._parent = self.get_parent()
        self.get_parent().add_child(copied_self)
        return copied_self

    def get_attached_elements(self):
        output = []
        for leaf in self.iterate_leaves():
            output.extend(leaf.content.xml_elements)
        return output

    def get_leaves(self, function=None):
        if isinstance(self.content, XSDElement):
            if function:
                return function(self)
            else:
                return self.content
        elif isinstance(self.content, XSDGroup):
            return self.get_children()[0].get_leaves(function=function)
        elif isinstance(self.content, XSDSequence) or isinstance(self.content, XSDChoice):
            output = [node.get_leaves(function=function) for node in self.get_children() if
                      node.get_leaves(function=function)]
            try:
                if not output or set(output) == {None}:
                    return None
            except TypeError:
                pass
            if len(output) == 1:
                return output[0]
            return output if isinstance(self.content, XSDSequence) else tuple(output)
        else:
            raise NotImplementedError

    def get_parent_xml_element(self):
        """
        :return: :obj:`~musicxml.xmlelement.xmlelement.XMLElement` which child container is attached to.
        """
        return self._parent_xml_element

    def get_required_element_names(self, intelligent_choice=False):
        def func(leaf):
            if leaf.requirements_fulfilled is False:
                return convert_to_xml_class_name(leaf.content.name)

            elif leaf.min_occurrences != 0 and False in [choice.requirements_fulfilled for choice in
                                                         leaf.choices_in_reversed_path]:
                if isinstance(leaf.get_parent().content,
                              XSDSequence) and leaf.get_parent().min_occurrences == 0 and not leaf.get_parent(

                ).force_validate:
                    pass
                else:
                    return convert_to_xml_class_name(leaf.content.name)

        self.check_required_elements(intelligent_choice)
        return self.get_leaves(func)

    def set_force_validate(self, node, val):
        self._force_validate = val
        for child in [ch for ch in self.get_children() if ch != node]:
            for n in child.traverse():
                if isinstance(n.content, XSDChoice):
                    if n.min_occurrences != 0 and not n.chosen_child and [ch for ch in n.get_children() if
                                                                          ch.min_occurrences != 0]:
                        n.requirements_fulfilled = False
                    break
                if isinstance(n.content, XSDSequence) and n.min_occurrences == 0:
                    break
                if isinstance(n.content, XSDSequence) and n.min_occurrences != 0 and not (
                        isinstance(n.get_parent().content, XSDGroup) and
                        n.get_parent().min_occurrences == 0):
                    n._force_validate = val

    def __repr__(self):
        return f"XMLChildContainer:{self.compact_repr} {self.get_position_in_tree()}"

    def __str__(self):
        return self.compact_repr
    
    def __copy__(self):
        copied = self.__class__(content=self.content.__copy__(), min_occurrences=self.min_occurrences,
                                max_occurrences=self.max_occurrences,
                                populate_children=False)
        for child in self.get_children():
            copied.add_child(child.__copy__())

        return copied


class XMLChildContainerFactory:
    def __init__(self, complex_type):
        self._child_container = None
        self._create_child_container(complex_type)

    def _create_child_container(self, complex_type):
        if 'XSDComplexType' not in [cls.__name__ for cls in complex_type.__mro__]:
            raise TypeError
        if not complex_type.get_xsd_indicator():
            raise XMLChildContainerFactoryError(f'complex_type {complex_type} has no xsd_indicator.')
        child_container = XMLChildContainer(*complex_type.get_xsd_indicator())
        self._child_container = child_container

    def get_child_container(self):
        return self._child_container



import xml.etree.ElementTree as ET
from unittest import TestCase

from musicxml.xsd.xsdindicator import XSDSequence
from musicxml.xsd.xsdtree import XSDTree


class TestSequence(TestCase):
    def setUp(self) -> None:
        xsd = """
                <xs:sequence xmlns:xs="http://www.w3.org/2001/XMLSchema">
                    <xs:element name="midi-channel" type="midi-16" minOccurs="0">
                        <xs:annotation>
                            <xs:documentation>The midi-channel element specifies a MIDI 1.0 channel numbers ranging from 1 to 16.</xs:documentation>
                        </xs:annotation>
                    </xs:element>
                    <xs:element name="midi-name" type="xs:string" minOccurs="0">
                        <xs:annotation>
                            <xs:documentation>The midi-name element corresponds to a ProgramName meta-event within a Standard MIDI File.</xs:documentation>
                        </xs:annotation>
                    </xs:element>
                    <xs:element name="midi-bank" type="midi-16384" minOccurs="0">
                        <xs:annotation>
                            <xs:documentation>The midi-bank element specifies a MIDI 1.0 bank number ranging from 1 to 16,384.</xs:documentation>
                        </xs:annotation>
                    </xs:element>
                    <xs:element name="midi-program" type="midi-128" minOccurs="0">
                        <xs:annotation>
                            <xs:documentation>The midi-program element specifies a MIDI 1.0 program number ranging from 1 to 128.</xs:documentation>
                        </xs:annotation>
                    </xs:element>
                    <xs:element name="midi-unpitched" type="midi-128" minOccurs="0">
                        <xs:annotation>
                            <xs:documentation>For unpitched instruments, the midi-unpitched element specifies a MIDI 1.0 note number ranging from 1 to 128. It is usually used with MIDI banks for percussion. Note that MIDI 1.0 note numbers are generally specified from 0 to 127 rather than the 1 to 128 numbering used in this element.</xs:documentation>
                        </xs:annotation>
                    </xs:element>
                    <xs:element name="volume" type="percent" minOccurs="0">
                        <xs:annotation>
                            <xs:documentation>The volume element value is a percentage of the maximum ranging from 0 to 100, with decimal values allowed. This corresponds to a scaling value for the MIDI 1.0 channel volume controller.</xs:documentation>
                        </xs:annotation>
                    </xs:element>
                    <xs:element name="pan" type="rotation-degrees" minOccurs="0">
                        <xs:annotation>
                            <xs:documentation>The pan and elevation elements allow placing of sound in a 3-D space relative to the listener. Both are expressed in degrees ranging from -180 to 180. For pan, 0 is straight ahead, -90 is hard left, 90 is hard right, and -180 and 180 are directly behind the listener.</xs:documentation>
                        </xs:annotation>
                    </xs:element>
                    <xs:element name="elevation" type="rotation-degrees" minOccurs="0">
                        <xs:annotation>
                            <xs:documentation>The elevation and pan elements allow placing of sound in a 3-D space relative to the listener. Both are expressed in degrees ranging from -180 to 180. For elevation, 0 is level with the listener, 90 is directly above, and -90 is directly below.</xs:documentation>
                        </xs:annotation>
                    </xs:element>
                </xs:sequence>
        """
        et = ET.fromstring(xsd)
        sequence_XSD_TREE = XSDTree(et)
        self.sequence = XSDSequence(sequence_XSD_TREE)

    def test_sequence_elements(self):
        assert self.sequence.elements == [('XMLMidiChannel', '0', '1'), ('XMLMidiName', '0', '1'), ('XMLMidiBank', '0', '1'),
                                          ('XMLMidiProgram', '0', '1'), ('XMLMidiUnpitched', '0', '1'), ('XMLVolume', '0', '1'),
                                          ('XMLPan', '0', '1'), ('XMLElevation', '0', '1')]

    def test_sequence_min_occurrences(self):
        xsd = """
                <xs:sequence xmlns:xs="http://www.w3.org/2001/XMLSchema">
                    <xs:element name="step" type="step"/>
                    <xs:element name="alter" type="semitones" minOccurs="0"/>
                    <xs:element name="octave" type="octave"/>
                </xs:sequence>
        """
        et = ET.fromstring(xsd)
        sequence_XSD_TREE = XSDTree(et)
        sequence = XSDSequence(sequence_XSD_TREE)
        assert sequence.elements == [('XMLStep', '1', '1'), ('XMLAlter', '0', '1'), ('XMLOctave', '1', '1')]

    def test_sequence_group_elements(self):
        xsd = """
        <xs:sequence xmlns:xs="http://www.w3.org/2001/XMLSchema">
            <xs:element name="group-name" type="group-name" minOccurs="0"/>
            <xs:element name="group-name-display" type="name-display" minOccurs="0">
                <xs:annotation>
                    <xs:documentation>Formatting specified in the group-name-display element overrides formatting specified in the group-name element.</xs:documentation>
                </xs:annotation>
            </xs:element>
            <xs:element name="group-abbreviation" type="group-name" minOccurs="0"/>
            <xs:element name="group-abbreviation-display" type="name-display" minOccurs="0">
                <xs:annotation>
                    <xs:documentation>Formatting specified in the group-abbreviation-display element overrides formatting specified in the group-abbreviation element.</xs:documentation>
                </xs:annotation>
            </xs:element>
            <xs:element name="group-symbol" type="group-symbol" minOccurs="0"/>
            <xs:element name="group-barline" type="group-barline" minOccurs="0"/>
            <xs:element name="group-time" type="empty" minOccurs="0">
                <xs:annotation>
                    <xs:documentation>The group-time element indicates that the displayed time signatures should stretch across all parts and staves in the group.</xs:documentation>
                </xs:annotation>
            </xs:element>
            <xs:group ref="editorial"/>
        </xs:sequence>
                """
        et = ET.fromstring(xsd)
        sequence_XSD_TREE = XSDTree(et)
        sequence = XSDSequence(sequence_XSD_TREE)
        assert sequence.elements == [('XMLGroupName', '0', '1'), ('XMLGroupNameDisplay', '0', '1'), ('XMLGroupAbbreviation', '0', '1'),
                                     ('XMLGroupAbbreviationDisplay', '0', '1'), ('XMLGroupSymbol', '0', '1'), ('XMLGroupBarline', '0', '1'),
                                     ('XMLGroupTime', '0', '1'),
                                     ('XMLFootnote', '0', '1'), ('XMLLevel', '0', '1')]



from unittest import TestCase

from musicxml.xsd.xsdindicator import *


class TestXSDGroups(TestCase):

    def test_xsd_group_doc(self):
        assert XSDGroupEditorial.__doc__ == 'The editorial group specifies editorial information for a musical element.'

    def test_xsd_group_get_sequence(self):
        assert XSDGroupEditorial().sequence.elements == [('XMLFootnote', '0', '1'), ('XMLLevel', '0', '1')]

    def test_xsd_group_score_header_sequence(self):
        assert XSDGroupScoreHeader().sequence.elements == [('XMLWork', '0', '1'), ('XMLMovementNumber', '0', '1'), ('XMLMovementTitle',
                                                                                                                    '0', '1'),
                                                           ('XMLIdentification', '0', '1'), ('XMLDefaults', '0', '1'),
                                                           ('XMLCredit', '0', 'unbounded'), ('XMLPartList', '1', '1')]



from unittest import TestCase

from musicxml.xmlelement.xmlelement import *


class TestXMLElementError(TestCase):
    # def test_simple_type_wrong_type(self):
    #     with self.assertRaises(TypeError) as err:
    #         XMLVoice()
    #     assert err.exception.args[0] == "XMLVoice's value 'None' can only be of types ['str'] not NoneType."

    def test_simple_type_wrong_type_with_enumeration(self):
        with self.assertRaises(ValueError) as err:
            XMLStep()
        assert err.exception.args[
                   0] == "XMLStep.value '' must be in ['A', 'B', 'C', 'D', 'E', 'F', 'G']"

    def test_simple_type_value_error_enumeration(self):
        with self.assertRaises(ValueError) as err:
            XMLFermata('something')
        assert err.exception.args[
                   0] == "XMLFermata: XSDSimpleTypeFermataShape.value 'something' must be in ['normal', 'angled', 'square', " \
                         "'double-angled', 'double-square', 'double-dot', 'half-curve', 'curlew', '']"

    def test_simple_type_value_error_restriction_min_exclusive(self):
        with self.assertRaises(ValueError) as err:
            XMLDivisions(-1)
        assert err.exception.args[0] == "XMLDivisions.value '-1' must be greater than '0'"

    def test_simple_type_value_error_restriction_min_inclusive(self):
        with self.assertRaises(ValueError) as err:
            XMLStaffSize(-1)
        assert err.exception.args[0] == "XMLStaffSize: XSDSimpleTypeNonNegativeDecimal.value '-1' must be greater than or equal to '0'"

    def test_simple_type_value_error_restriction_max_inclusive(self):
        with self.assertRaises(ValueError) as err:
            XMLPan(190)
        assert err.exception.args[0] == "XMLPan.value '190' must be less than or equal to '180'"

    def test_simple_type_attribute_value_err_pattern(self):
        w = XMLTie(type='start')
        with self.assertRaises(ValueError) as err:
            w.time_only = 'bla'
        assert err.exception.args[0] == "XSDSimpleTypeTimeOnly.value 'bla' must match the following pattern: [1-9][0-9]*(, ?[1-9][0-9]*)*"

    def test_simple_type_value_error_union(self):
        fs = XMLSymbol('bal')
        with self.assertRaises(TypeError) as err:
            fs.letter_spacing = 'not normal'
        assert err.exception.args[
                   0] == "XSDSimpleTypeNumberOrNormal's value 'not normal' can only be of types ['float', 'int'] not str. XSDSimpleTypeNumberOrNormal.value can also be ['normal']"



import re
from unittest import TestCase

from musicxml.tests.util import MusicXmlTestCase
from musicxml.util.core import get_cleaned_token, convert_to_xml_class_name, \
    replace_key_underline_with_hyphen
from musicxml.generate_classes.utils import get_simple_type_all_base_classes
from musicxml.util.helprervariables import name_character


class TestHelperFunctions(MusicXmlTestCase):

    def test_get_cleaned_token(self):
        expected = 'Hello Alfons'
        assert get_cleaned_token('Hello\nAlfons') == expected
        assert get_cleaned_token('Hello\tAlfons') == expected
        assert get_cleaned_token('Hello\rAlfons') == expected
        assert get_cleaned_token('Hello\t  Alfons') == expected
        assert get_cleaned_token('Hello    Alfons') == expected

    def test_replace_key_underline_with_hyphen(self):
        dict_ = {'a': 1, 'b-c': 2, 'e_f': 3}
        assert replace_key_underline_with_hyphen(dict_) == {'a': 1, 'b-c': 2, 'e-f': 3}
        dict_ = {'a': 1, 'b-c': 2, 'b_c': 3}
        with self.assertRaises(KeyError):
            replace_key_underline_with_hyphen(dict_)


class TestHelperVariables(TestCase):
    def test_name_character(self):
        pattern = rf"({name_character})+"
        p = re.compile(pattern)
        assert p.fullmatch('HeL1lo') is not None
        assert p.fullmatch('HeL1lo@') is None
        assert p.fullmatch('HeL1 lo') is None
        assert p.fullmatch('HeL1.o') is not None
        assert p.fullmatch('HeL1:._-o') is not None
        assert p.fullmatch('ÖÜöüäÄ:._-o') is not None
        assert p.fullmatch('&:._-o') is None


class TestConvertToXMLClassName(TestCase):
    def test_conversion(self):
        assert convert_to_xml_class_name('something') == 'XMLSomething'
        assert convert_to_xml_class_name('something-some-thing') == 'XMLSomethingSomeThing'



import copy
import xml.etree.ElementTree as ET
from contextlib import redirect_stdout
from pathlib import Path
from string import Template

from musicxml.util.core import convert_to_xml_class_name, convert_to_xsd_class_name
from musicxml.xmlelement.containers import containers
from musicxml.xsd.xsdtree import XSDTree, XSD_TREE_DICT
from musicxml.xsd.xsdcomplextype import *
from musicxml.xsd.xsdcomplextype import __all__ as all_complex_types
from musicxml.xsd.xsdsimpletype import *
from musicxml.xsd.xsdsimpletype import __all__ as all_simple_types

default_path = Path(__file__).parent / 'defaults' / 'xmlelement.py'
target_path = Path(__file__).parent.parent / 'xmlelement' / 'xmlelement.py'

template_string = """
class $class_name($base_classes):
    \"\"\"
    $doc
    \"\"\"
    
    TYPE = $xsd_type
    _SEARCH_FOR_ELEMENT = $search_for
    XSD_TREE = XSD_TREE_DICT['element'].get('$name')
"""

typed_elements = set(
    (item[0], item[1].type) for item in XSD_TREE_DICT['element'].items() if item[1].type is not None
)

typed_elements.add(('score-partwise', 'score-partwise'))
typed_elements.add(('part', 'part'))
typed_elements.add(('measure', 'measure'))
typed_elements.add(('directive', 'directive'))

typed_elements = sorted(typed_elements)


def generate_child_parent_dict() -> dict:
    """
    :return: a dictionary with name of a child :obj:`~musicxml.xmlelement.xmlelement.:obj:`~musicxml.xmlelement.xmlelement.XMLElement`` as key assiciated with a tuple of possible :obj:`~musicxml.xmlelement.xmlelement.:obj:`~musicxml.xmlelement.xmlelement.XMLElement`` parent names.

    >>> child_parent_dict = generate_child_parent_dict()
    >>> child_parent_dict['XMLStaccato']
    {'XMLArticulations'}
    """
    output = {}
    for name, type_ in typed_elements:
        try:
            class_name = convert_to_xml_class_name(name)
            type_name = convert_to_xsd_class_name(type_, type_='complex_type')
            container = containers[type_name]
            elements = sorted(set(convert_to_xml_class_name(l.content.name) for l in container.iterate_leaves()))
            for el in elements:
                if output.get(el):
                    output[el].add(class_name)
                else:
                    output[el] = {class_name}
        except (KeyError, ValueError):
            pass
    return output


child_parent_dict = generate_child_parent_dict()

extra_classes = {
    'score-partwise':
        {'search_for': ".//{*}element[@name='score-partwise']",
         'xsd_type': 'XSDComplexTypeScorePartwise',
         },
    'part':
        {'search_for': ".//{*}element[@name='score-partwise']//{*}element[@name='part']",
         'xsd_type': 'XSDComplexTypePart'
         },
    'measure':
        {'search_for': ".//{*}element[@name='score-partwise']//{*}element[@name='measure']",
         'xsd_type': 'XSDComplexTypeMeasure'
         },
    'directive':
        {'search_for': ".//{*}complexType[@name='attributes']//{*}element[@name='directive']",
         'xsd_type': 'XSDComplexTypeDirective'

         }
}

xml_element_class_names = []


def element_class_as_string(element_name_type):
    def get_doc():
        def get_external_doc_link():

            return f"`external documentation <https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/{element_name_type[0]}/>`_"

        def get_attributes_doc():
            output = ""
            try:
                possible_attributes = eval(xsd_type).get_xsd_attributes()
                string_possible_attributes = []
                for att in possible_attributes:
                    repr_ = ''
                    if att.name:
                        repr_ += f"``{'_'.join(att.name.split('-'))}``"
                    if att.type_:
                        try:
                            repr_ += f"\\@ :obj:`~musicxml.xsd.xsdsimpletype.{att.type_.__name__}`"
                        except AttributeError:
                            breakpoint()
                            pass
                    if att.is_required:
                        repr_ += '\\@required'

                    if repr_ != '':
                        string_possible_attributes.append(repr_)

                if string_possible_attributes:
                    output += '    ``Possible attributes``: '
                    output += f"{', '.join(sorted(string_possible_attributes))}"

            except (AttributeError, KeyError, NameError):
                pass
            return output

        def get_possible_parents():
            parent_names = child_parent_dict.get(class_name)
            if parent_names is None:
                return """
    .. todo::         
       Possible parents
"""
            possible_parents = ", ".join(sorted(set(f":obj:`~{parent_name}`" for parent_name in parent_names)))
            output = "``Possible parents``:"
            output += f"{possible_parents}"
            return output

        def get_possible_children(container):
            possible_children = ", ".join(sorted(set(f":obj:`~{convert_to_xml_class_name(l.content.name)}`" for l in
                                                     container.iterate_leaves())))

            output = '    ``Possible children``:'
            output += f"    {possible_children}"
            return output

        output = get_external_doc_link()
        if xsd_tree.get_doc():
            output += '\n'
            output += '\n'
            output += xsd_tree.get_doc()
        output += '\n'
        output += '\n'
        if xsd_type in all_complex_types:
            complex_type_doc = eval(xsd_type).__doc__

            if complex_type_doc:
                if complex_type_doc:
                    output += '\n'
                    output += '\n'
                output += '``complexType``: '
                output += complex_type_doc
                if output.count('\n') > 1:
                    output = output.replace('\n', '\n    ')
            if get_attributes_doc() != '':
                output += '\n'
                output += '\n'
                output += get_attributes_doc()
            try:
                container = containers[xsd_type]
                container_tree_representation = copy.copy(container).get_tree_representation(tab=lambda x: (x.level * '    ') + '       ')
                container_tree_representation = container_tree_representation[:-1]
                if output != "":
                    output += '\n'
                    output += '\n'

                output += get_possible_children(container)
                output += '\n'
                output += '\n'

                output += "    ``XSD structure:``\n"
                output += '\n'
                output += "    .. code-block::\n"
                output += '\n'
                output += container_tree_representation

            except KeyError:
                pass
        elif xsd_type in all_simple_types:
            simple_type_doc = eval(xsd_type).__doc__
            if simple_type_doc:
                if simple_type_doc and output and output != "":
                    output += '\n'
                    output += '\n'
                output += '``simpleType``: '
                output += simple_type_doc
                if output.count('\n') > 1:
                    output = output.replace('\n', '\n    ')
        else:
            pass
        if element_name_type[0] != 'score-partwise':
            output += '\n'
            output += '\n'
            output += get_possible_parents()
        return output

    search_for = f'"{extra_classes[element_name_type[0]]["search_for"]}"' if extra_classes.get(element_name_type[0]) else None
    name = element_name_type[0]
    xsd_tree = XSD_TREE_DICT['element'][name]
    class_name = convert_to_xml_class_name(name)
    xml_element_class_names.append(class_name)

    xsd_type = extra_classes[element_name_type[0]]['xsd_type'] if extra_classes.get(element_name_type[0]) else None
    if not xsd_type:
        try:
            xsd_type = convert_to_xsd_class_name(xsd_tree.type, 'complex_type')
            if xsd_type not in all_complex_types:
                raise ValueError
        except ValueError:
            xsd_type = convert_to_xsd_class_name(xsd_tree.type, 'simple_type')
    base_classes = ('XMLElement',)

    t = Template(template_string).substitute(class_name=class_name, base_classes=', '.join(base_classes), xsd_type=xsd_type,
                                             name=name, search_for=search_for, doc=get_doc())
    if element_name_type[0] == 'score-partwise':
        t += '\n'
        t += """    def write(self, path: 'pathlib.Path', intelligent_choice: bool=False) -> None:
        \"\"\"
        :param path: Output xml file path, required.
        :param intelligent_choice: Set to True if you wish to use intelligent choice in final checks to be able to change the attachment 
                                   order of XMLElement children in self.child_container_tree if an Exception was thrown and other choices 
                                   can still be checked. (No GUARANTEE!)
        :return: None
        \"\"\"
        with open(path, 'w') as file:
            file.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?>\\n')
            file.write(self.to_string(intelligent_choice=intelligent_choice))
"""
    return t


# typed_elements.remove(('senza-misura', 'xs:string'))

with open(target_path, 'w+') as f:
    with open(default_path, 'r') as default:
        with redirect_stdout(f):
            print(default.read())
    with redirect_stdout(f):
        for element_name_type in typed_elements:
            print(element_class_as_string(element_name_type))
        print(f'__all__={xml_element_class_names}')
